# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
#
#  FreeType high-level python API - Copyright 2011-2015 Nicolas P. Rougier
#  Distributed under the terms of the new BSD license.
#
# -----------------------------------------------------------------------------
"""
A list of bit flags used in the 'face_flags' field of the FT_FaceRec
structure. They inform client applications of properties of the corresponding
face.


FT_FACE_FLAG_SCALABLE

  Indicates that the face contains outline glyphs. This doesn't prevent bitmap
  strikes, i.e., a face can have both this and and FT_FACE_FLAG_FIXED_SIZES
  set.


FT_FACE_FLAG_FIXED_SIZES

  Indicates that the face contains bitmap strikes. See also the
  'num_fixed_sizes' and 'available_sizes' fields of FT_FaceRec.


FT_FACE_FLAG_FIXED_WIDTH

  Indicates that the face contains fixed-width characters (like Courier,
  Lucido, MonoType, etc.).


FT_FACE_FLAG_SFNT

  Indicates that the face uses the 'sfnt' storage scheme. For now, this means
  TrueType and OpenType.


FT_FACE_FLAG_HORIZONTAL

  Indicates that the face contains horizontal glyph metrics. This should be set
  for all common formats.


FT_FACE_FLAG_VERTICAL

  Indicates that the face contains vertical glyph metrics. This is only
  available in some formats, not all of them.


FT_FACE_FLAG_KERNING

  Indicates that the face contains kerning information. If set, the kerning
  distance can be retrieved through the function FT_Get_Kerning. Otherwise the
  function always return the vector (0,0). Note that FreeType doesn't handle
  kerning data from the 'GPOS' table (as present in some OpenType fonts).


FT_FACE_FLAG_MULTIPLE_MASTERS

  Indicates that the font contains multiple masters and is capable of
  interpolating between them. See the multiple-masters specific API for
  details.


FT_FACE_FLAG_GLYPH_NAMES

  Indicates that the font contains glyph names that can be retrieved through
  FT_Get_Glyph_Name. Note that some TrueType fonts contain broken glyph name
  tables. Use the function FT_Has_PS_Glyph_Names when needed.


FT_FACE_FLAG_EXTERNAL_STREAM

  Used internally by FreeType to indicate that a face's stream was provided by
  the client application and should not be destroyed when FT_Done_Face is
  called. Don't read or test this flag.


FT_FACE_FLAG_HINTER

  Set if the font driver has a hinting machine of its own. For example, with
  TrueType fonts, it makes sense to use data from the SFNT 'gasp' table only if
  the native TrueType hinting engine (with the bytecode interpreter) is
  available and active.


FT_FACE_FLAG_CID_KEYED

  Set if the font is CID-keyed. In that case, the font is not accessed by glyph
  indices but by CID values. For subsetted CID-keyed fonts this has the
  consequence that not all index values are a valid argument to
  FT_Load_Glyph. Only the CID values for which corresponding glyphs in the
  subsetted font exist make FT_Load_Glyph return successfully; in all other
  cases you get an 'FT_Err_Invalid_Argument' error.

  Note that CID-keyed fonts which are in an SFNT wrapper don't have this flag
  set since the glyphs are accessed in the normal way (using contiguous
  indices); the 'CID-ness' isn't visible to the application.


FT_FACE_FLAG_TRICKY

  Set if the font is 'tricky', this is, it always needs the font format's
  native hinting engine to get a reasonable result. A typical example is the
  Chinese font 'mingli.ttf' which uses TrueType bytecode instructions to move
  and scale all of its subglyphs.

  It is not possible to autohint such fonts using FT_LOAD_FORCE_AUTOHINT; it
  will also ignore FT_LOAD_NO_HINTING. You have to set both FT_LOAD_NO_HINTING
  and FT_LOAD_NO_AUTOHINT to really disable hinting; however, you probably
  never want this except for demonstration purposes.

  Currently, there are six TrueType fonts in the list of tricky fonts; they are
  hard-coded in file 'ttobjs.c'.
"""
FT_FACE_FLAGS = { 'FT_FACE_FLAG_SCALABLE'          : 1 <<  0,
                  'FT_FACE_FLAG_FIXED_SIZES'       : 1 <<  1,
                  'FT_FACE_FLAG_FIXED_WIDTH'       : 1 <<  2,
                  'FT_FACE_FLAG_SFNT'              : 1 <<  3,
                  'FT_FACE_FLAG_HORIZONTAL'        : 1 <<  4,
                  'FT_FACE_FLAG_VERTICAL'          : 1 <<  5,
                  'FT_FACE_FLAG_KERNING'           : 1 <<  6,
                  'FT_FACE_FLAG_FAST_GLYPHS'       : 1 <<  7,
                  'FT_FACE_FLAG_MULTIPLE_MASTERS'  : 1 <<  8,
                  'FT_FACE_FLAG_GLYPH_NAMES'       : 1 <<  9,
                  'FT_FACE_FLAG_EXTERNAL_STREAM'   : 1 << 10,
                  'FT_FACE_FLAG_HINTER'            : 1 << 11,
                  'FT_FACE_FLAG_CID_KEYED'         : 1 << 12,
                  'FT_FACE_FLAG_TRICKY'            : 1 << 13
}
globals().update(FT_FACE_FLAGS)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
#
#  FreeType high-level python API - Copyright 2011-2015 Nicolas P. Rougier
#  Distributed under the terms of the new BSD license.
#
# -----------------------------------------------------------------------------
import numpy
import os, sys
from freetype import *


class ColorMap:
    ''' A colormap is used to map scalar values to colors. It is build by
        adding couples of (value,color) where value must be between 0 and 1.
        The 'scale' method allows to specify the range of the colormap and
        the 'color' method then returns a color for any value. '''

    def __init__ (self, colors):
        self.colors = colors
        self.min    = 0
        self.max    = 1

    def scale (self, min, max):
        self.min, self.max = min,max

    def color (self, value):
        ''' Return the color corresponding to value. '''
        if not len(self.colors):
            return (0,0,0)
        elif len(self.colors) == 1:
            return self.colors[0][1]
        elif value < self.min:
            return self.colors[0][1]
        elif value > self.max:
            return self.colors[-1][1]
        value = (value-self.min)/(self.max-self.min)
        sup_color = self.colors[0]
        inf_color = self.colors[-1]
        for i in range (len(self.colors)-1):
            if value < self.colors[i+1][0]:
                inf_color = self.colors[i]
                sup_color = self.colors[i+1]
                break
        r = (value-inf_color[0]) / (sup_color[0] - inf_color[0])
        if r < 0: r = -r
        color = [sup_color[1][0]*r + inf_color[1][0]*(1-r),
                 sup_color[1][1]*r + inf_color[1][1]*(1-r),
                 sup_color[1][2]*r + inf_color[1][2]*(1-r)]
        return color

# Some colormaps
CM_IceAndFire = ColorMap([(0.00, (0.0, 0.0, 1.0)),
                         (0.25, (0.0, 0.5, 1.0)),
                         (0.50, (1.0, 1.0, 1.0)),
                         (0.75, (1.0, 1.0, 0.0)),
                         (1.00, (1.0, 0.0, 0.0))])
CM_Ice = ColorMap([(0.00, (0.0, 0.0, 1.0)),
                   (0.50, (0.5, 0.5, 1.0)),
                   (1.00, (1.0, 1.0, 1.0))])
CM_Fire = ColorMap([(0.00, (1.0, 1.0, 1.0)),
                    (0.50, (1.0, 1.0, 0.0)),
                    (1.00, (1.0, 0.0, 0.0))])
CM_Hot = ColorMap([(0.00, (0.0, 0.0, 0.0)),
                   (0.33, (1.0, 0.0, 0.0)),
                   (0.66, (1.0, 1.0, 0.0)),
                   (1.00, (1.0, 1.0, 1.0))])
CM_Grey = ColorMap([(0.00, (0.0, 0.0, 0.0)),
                    (1.00, (1.0, 1.0, 1.0))])



def imshow (Z, vmin=None, vmax=None, cmap=CM_Hot, show_cmap=False):
    ''' Show a 2D numpy array using terminal colors '''

    if len(Z.shape) != 2:
        print ("Cannot display non 2D array")
        return

    vmin = vmin or Z.min()
    vmax = vmax or Z.max()
    cmap.scale (vmin, vmax)

    # Build initialization string that setup terminal colors
    init = ''
    for i in range(240):
        v = cmap.min + (i/240.0)* (cmap.max - cmap.min)
        r,g,b = cmap.color (v)
        init += "\x1b]4;%d;rgb:%02x/%02x/%02x\x1b\\" % (16+i, int(r*255),int(g*255),int(b*255))

    # Build array data string
    data = ''
    for i in range(Z.shape[0]):
        for j in range(Z.shape[1]):
            c = 16 + int( ((Z[Z.shape[0]-i-1,j]-cmap.min) / (cmap.max-cmap.min))*239)
            if (c < 16):
                c=16
            elif (c > 255):
                c=255
            data += "\x1b[48;5;%dm  " % c
            u = cmap.max - (i/float(Z.shape[0]-1)) * ((cmap.max-cmap.min))
        if show_cmap:
            data += "\x1b[0m  "
            data += "\x1b[48;5;%dm  " % (16 + (1-i/float(Z.shape[0]))*239)
            data += "\x1b[0m %+.2f" % u
        data += "\x1b[0m\n"
    print (init+data[:-1]+'\x1b[0m')


if __name__ == '__main__':
    face = Face('./Vera.ttf')
    face.set_char_size( 32*64 )
    face.load_glyph(face.get_char_index('S'))
    slot = face.glyph
    bitmap = slot.bitmap
    data, rows, width = bitmap.buffer, bitmap.rows, bitmap.width
    Z = numpy.array(data,dtype=float).reshape(rows,width)
    Z = Z[::-1,:]
    imshow (Z, cmap=CM_Grey)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (C) 2009-2010  Nicolas P. Rougier
#
# Distributed under the terms of the BSD License. The full license is in
# the file COPYING, distributed as part of this software.
# -----------------------------------------------------------------------------
#
# Copyright Tristam Macdonald 2008.
#
# Distributed under the Boost Software License, Version 1.0
# (see http://www.boost.org/LICENSE_1_0.txt)
#
''' Base shader class.

    Example:
    --------
    shader = Shader()

    shader.bind()
    glActiveTexture(GL_TEXTURE1)
    glBindTexture(lut.target, lut.id)
    shader.uniformi('lut', 1)

    glActiveTexture(GL_TEXTURE0)
    glBindTexture(texture.target, texture.id)
    shader.uniformi('texture', 0)
    shader.uniformf('pixel', 1.0/texture.width, 1.0/texture.height)

    texture.blit(x,y,w,h)
    shader.unbind()
'''
import os
import OpenGL.GL as gl
import ctypes

class Shader:
    ''' Base shader class. '''

    def __init__(self, vert = None, frag = None, name=''):
        ''' vert, frag and geom take arrays of source strings
            the arrays will be concatenated into one string by OpenGL.'''

        self.uniforms = {}
        self.name = name
        # create the program handle
        self.handle = gl.glCreateProgram()
        # we are not linked yet
        self.linked = False
        # create the vertex shader
        self._build_shader(vert, gl.GL_VERTEX_SHADER)
        # create the fragment shader
        self._build_shader(frag, gl.GL_FRAGMENT_SHADER)
        # the geometry shader will be the same, once pyglet supports the
        # extension self.createShader(frag, GL_GEOMETRY_SHADER_EXT) attempt to
        # link the program
        self._link()

    def _build_shader(self, strings, stype):
        ''' Actual building of the shader '''

        count = len(strings)
        # if we have no source code, ignore this shader
        if count < 1:
            return

        # create the shader handle
        shader = gl.glCreateShader(stype)

        # Upload shader code
        gl.glShaderSource(shader, strings)

        # compile the shader
        gl.glCompileShader(shader)

        # retrieve the compile status
        status = gl.glGetShaderiv(shader, gl.GL_COMPILE_STATUS)

        # if compilation failed, print the log
        if not status:
            # display the log
            print (gl.glGetShaderInfoLog(shader))
        else:
            # all is well, so attach the shader to the program
            gl.glAttachShader(self.handle, shader)

    def _link(self):
        ''' Link the program '''

        gl.glLinkProgram(self.handle)
        # retrieve the link status
        temp = ctypes.c_int(0)
        gl.glGetProgramiv(self.handle, gl.GL_LINK_STATUS, ctypes.byref(temp))

        # if linking failed, print the log
        if not temp:
            # retrieve the log length
            gl.glGetProgramiv(self.handle,
                              gl.GL_INFO_LOG_LENGTH, ctypes.byref(temp))
            # create a buffer for the log
            #buffer = ctypes.create_string_buffer(temp.value)
            # retrieve the log text
            log = gl.glGetProgramInfoLog(self.handle) #, temp, None, buffer)
            # print the log to the console
            print (log)
        else:
            # all is well, so we are linked
            self.linked = True

    def bind(self):
        ''' Bind the program, i.e. use it. '''
        gl.glUseProgram(self.handle)

    def unbind(self):
        ''' Unbind whatever program is currently bound - not necessarily this
            program, so this should probably be a class method instead. '''
        gl.glUseProgram(0)

    def uniformf(self, name, *vals):
        ''' Uploads float uniform(s), program must be currently bound. '''

        loc = self.uniforms.get(name,
                                gl.glGetUniformLocation(self.handle,name))
        self.uniforms[name] = loc

        # Check there are 1-4 values
        if len(vals) in range(1, 5):
            # Select the correct function
            { 1 : gl.glUniform1f,
              2 : gl.glUniform2f,
              3 : gl.glUniform3f,
              4 : gl.glUniform4f
              # Retrieve uniform location, and set it
            }[len(vals)](loc, *vals)

    def uniformi(self, name, *vals):
        ''' Upload integer uniform(s), program must be currently bound. '''

        loc = self.uniforms.get(name,
                                gl.glGetUniformLocation(self.handle,name))
        self.uniforms[name] = loc

        # Checks there are 1-4 values
        if len(vals) in range(1, 5):
            # Selects the correct function
            { 1 : gl.glUniform1i,
              2 : gl.glUniform2i,
              3 : gl.glUniform3i,
              4 : gl.glUniform4i
              # Retrieves uniform location, and set it
            }[len(vals)](loc, *vals)


    def uniform_matrixf(self, name, mat):
        ''' Upload uniform matrix, program must be currently bound. '''

        loc = self.uniforms.get(name,
                                gl.glGetUniformLocation(self.handle,name))
        self.uniforms[name] = loc

        # Upload the 4x4 floating point matrix
        gl.glUniformMatrix4fv(loc, 1, False, (ctypes.c_float * 16)(*mat))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  OT-SVG example, alternative version based on Skia
#
#  Copyright 2023 Hin-Tak Leung
#  Distributed under the terms of the new BSD license.

#  See also the other one, the rsvg-based OT-SVG example.

# "ot-svg-draw-skia.py" was written based on skia m87.
# This example requires m117, and also accelerated GL integration;
# should work on any OT-SVG font, unlike "ot-svg-draw-skia.py".

if __name__ == '__main__':
    import sys

    from ctypes import byref, pythonapi, cast, c_char_p
    from freetype import Face, get_handle, FT_Property_Set, FT_LOAD_COLOR, FT_LOAD_RENDER
    from OpenGL import GL
    import glfw
    import skia
    from skia import ImageInfo, ColorType, AlphaType

    from skia_ot_svg_module import hooks
    from skia_glfw_module import glfw_window, skia_surface

    execname = sys.argv[0]

    if len(sys.argv) < 2:
        print("Example usage: %s TrajanColor-Concept.otf" % execname)
        exit(1)

    face = Face(sys.argv[1])

    face.set_char_size( 160*64 )
    library = get_handle()

    FT_Property_Set( library, b"ot-svg", b"svg-hooks", byref(hooks) ) # python 3 only syntax
    face.load_char('A', FT_LOAD_COLOR | FT_LOAD_RENDER )

    bitmap = face.glyph.bitmap
    width = face.glyph.bitmap.width
    rows = face.glyph.bitmap.rows

    if ( face.glyph.bitmap.pitch != width * 4 ):
        raise RuntimeError('pitch != width * 4 for color bitmap: Please report this.')

    WIDTH, HEIGHT = 2*width, rows

    glyphBitmap = skia.Bitmap()
    glyphBitmap.setInfo(ImageInfo.Make(bitmap.width, bitmap.rows,
                                       ColorType.kBGRA_8888_ColorType,
                                       AlphaType.kPremul_AlphaType),
                        bitmap.pitch)
    glyphBitmap.setPixels(pythonapi.PyMemoryView_FromMemory(cast(bitmap._FT_Bitmap.buffer, c_char_p),
                                                            bitmap.rows * bitmap.pitch,
                                                            0x200), # Read-Write
                          )

    with glfw_window(WIDTH, HEIGHT) as window:
        GL.glClear(GL.GL_COLOR_BUFFER_BIT)

        with skia_surface(window) as surface:
            with surface as canvas:
                canvas.drawBitmap(glyphBitmap, 0, 0)
                canvas.drawBitmap(glyphBitmap, width/2, 0)
                canvas.drawBitmap(glyphBitmap, width, 0)
            surface.flushAndSubmit()
            glfw.swap_buffers(window)

            while (glfw.get_key(window, glfw.KEY_ESCAPE) != glfw.PRESS
                   and not glfw.window_should_close(window)):
                glfw.wait_events()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
#
#  pycairo/cairocffi-based glyph-mono/alpha example - Copyright 2017 Hin-Tak Leung
#  Distributed under the terms of the new BSD license.
#
#  rewrite of the numply,matplotlib-based example from Nicolas P. Rougier
#  - Not immitating the upside-downness of glyph-monochrome/glyph-alpha

#  This script default to normal(8-bit) rendering, but render to mono
#  if any argument is specified.
#
#  Mono rendering requires libtiff on small-endian platforms. See
#  comments in bitmap_to_surface.py.
#
# -----------------------------------------------------------------------------

'''
Glyph bitmap monochrome/alpha rendring
'''
from freetype import *

# use Matrix() from Cairo instead of from Freetype
from cairo import Context, ImageSurface, FORMAT_ARGB32, SurfacePattern, FILTER_BEST, Matrix
from bitmap_to_surface import make_image_surface

if __name__ == '__main__':
    from PIL import Image
    import sys

    face = Face('./Vera.ttf')
    face.set_char_size( 48*64 )

    if len(sys.argv) < 2:
        # Normal(8-bit) Rendering
        face.load_char('S', FT_LOAD_RENDER |
                       FT_LOAD_TARGET_NORMAL )
    else:
        # Mono(1-bit) Rendering
        face.load_char('S', FT_LOAD_RENDER |
                       FT_LOAD_TARGET_MONO )

    bitmap = face.glyph.bitmap
    width  = face.glyph.bitmap.width
    rows   = face.glyph.bitmap.rows
    pitch  = face.glyph.bitmap.pitch

    glyph_surface = make_image_surface(face.glyph.bitmap)

    surface = ImageSurface(FORMAT_ARGB32, 800, 600)
    ctx = Context(surface)
    ctx.rectangle(0,0,800,600)
    ctx.set_line_width(0)
    ctx.set_source_rgb (0.5 , 0.5, 0.5)
    ctx.fill()
    #
    scale = 480.0 / rows
    ctx.set_source_surface(glyph_surface, 0, 0)
    pattern = ctx.get_source()
    SurfacePattern.set_filter(pattern, FILTER_BEST)
    scalematrix = Matrix()
    scalematrix.scale(1.0/scale,1.0/scale)
    scalematrix.translate(-(400.0 - width *scale /2.0 ), -60)
    pattern.set_matrix(scalematrix)
    ctx.set_source_rgb (0 , 0, 1)
    ctx.mask(pattern)
    ctx.fill()
    surface.flush()
    surface.write_to_png("glyph-mono+alpha-cairo.png")
    Image.open("glyph-mono+alpha-cairo.png").show()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
#
#  FreeType high-level python API - Copyright 2011-2015 Nicolas P. Rougier
#  Distributed under the terms of the new BSD license.
#
# -----------------------------------------------------------------------------
'''
Basic outline wrapper test
'''
from freetype import *
import ctypes
import contextlib

@contextlib.contextmanager
def new_outline(n_points, n_contours):
    library = get_handle()
    raw_ft_outline = FT_Outline()
    err = FT_Outline_New(
        library, FT_UInt(n_points), FT_Int(n_contours),
        ctypes.byref(raw_ft_outline)
    )
    if err: raise FT_Exception(err)
    try:
        raw_ft_outline.n_points   = 0
        raw_ft_outline.n_contours = 0
        yield Outline(raw_ft_outline)
    finally:
        FT_Outline_Done(library, ctypes.byref(raw_ft_outline))

if __name__ == '__main__':
    face = Face('./Vera.ttf')
    face.set_char_size( 4*48*64 )
    flags = FT_LOAD_DEFAULT | FT_LOAD_NO_BITMAP
    face.load_char('S', flags )
    slot = face.glyph
    outline = slot.outline

    bbox = outline.get_bbox()
    cbox = outline.get_cbox()
    assert (bbox.xMin, bbox.yMin, bbox.xMax, bbox.yMax) == \
        (810, -192, 7116, 9152)
    assert (cbox.xMin, cbox.yMin, cbox.xMax, cbox.yMax) == \
        (810, -192, 7116, 9152)
    assert outline.get_outside_border() == 0
    assert outline.get_inside_border() == 1
    assert len(outline.contours) == 1
    assert len(outline.points) == 40
    assert len(outline.tags) == 40

    stroker = Stroker()
    stroker.set(64, FT_STROKER_LINECAP_ROUND, FT_STROKER_LINEJOIN_ROUND, 0)
    stroker.parse_outline(outline, False)

    n_points, n_contours = stroker.get_counts()
    with new_outline(n_points, n_contours) as stroked_outline:
        stroker.export(stroked_outline)
        bbox = stroked_outline.get_bbox()
        cbox = stroked_outline.get_cbox()
        assert (bbox.xMin, bbox.yMin, bbox.xMax, bbox.yMax) == \
            (746, -256, 7180, 9216)
        assert (cbox.xMin, cbox.yMin, cbox.xMax, cbox.yMax) == \
            (746, -256, 7180, 9216)
        assert stroked_outline.get_outside_border() == 0
        assert stroked_outline.get_inside_border() == 1
        assert len(stroked_outline.contours) == 2
        assert len(stroked_outline.points) == 225
        assert len(stroked_outline.tags) == 225

    border = outline.get_outside_border()
    n_points, n_contours = stroker.get_border_counts(border)
    with new_outline(n_points, n_contours) as outer_outline:
        stroker.export_border(border, outer_outline)
        bbox = outer_outline.get_bbox()
        cbox = outer_outline.get_cbox()
        assert (bbox.xMin, bbox.yMin, bbox.xMax, bbox.yMax) == \
            (746, -256, 7180, 9216)
        assert (cbox.xMin, cbox.yMin, cbox.xMax, cbox.yMax) == \
            (746, -256, 7180, 9216)
        assert outer_outline.get_outside_border() == 0
        assert outer_outline.get_inside_border() == 1
        assert len(outer_outline.contours) == 1
        assert len(outer_outline.points) == 121
        assert len(outer_outline.tags) == 121

    border = outline.get_inside_border()
    n_points, n_contours = stroker.get_border_counts(border)
    with new_outline(n_points, n_contours) as inner_outline:
        stroker.export_border(border, inner_outline)
        bbox = inner_outline.get_bbox()
        cbox = inner_outline.get_cbox()
        assert (bbox.xMin, bbox.yMin, bbox.xMax, bbox.yMax) == \
            (813, -128, 7052, 9088)
        assert (cbox.xMin, cbox.yMin, cbox.xMax, cbox.yMax) == \
            (813, -128, 7052, 9088)
        assert inner_outline.get_outside_border() == 1
        assert inner_outline.get_inside_border() == 0
        assert len(inner_outline.contours) == 1
        assert len(inner_outline.points) == 104
        assert len(inner_outline.tags) == 104



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
#
#  FreeType high-level python API - Copyright 2011-2015 Nicolas P. Rougier
#  Distributed under the terms of the new BSD license.
#
# -----------------------------------------------------------------------------
"""
A list of valid values for the 'encoding_id' for TT_PLATFORM_ADOBE
charmaps. This is a FreeType-specific extension!

TT_ADOBE_ID_STANDARD

  Adobe standard encoding.


TT_ADOBE_ID_EXPERT

  Adobe expert encoding.


TT_ADOBE_ID_CUSTOM

  Adobe custom encoding.


TT_ADOBE_ID_LATIN_1

  Adobe Latin 1 encoding.
"""

TT_ADOBE_IDS = {
    'TT_ADOBE_ID_STANDARD' : 0,
    'TT_ADOBE_ID_EXPERT'   : 1,
    'TT_ADOBE_ID_CUSTOM'   : 2,
    'TT_ADOBE_ID_LATIN_1'  : 3 }
globals().update(TT_ADOBE_IDS)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
#
#  FreeType high-level python API - Copyright 2011-2015 Nicolas P. Rougier
#  Distributed under the terms of the new BSD license.
#
# -----------------------------------------------------------------------------
'''
Glyph bitmap monochrome rendring
'''
from freetype import *

def bits(x):
    data = []
    for i in range(8):
        data.insert(0, int((x & 1) == 1))
        x = x >> 1
    return data

if __name__ == '__main__':
    import numpy
    import matplotlib.pyplot as plt

    face = Face('./Vera.ttf')
    face.set_char_size( 48*64 )
    face.load_char('S', FT_LOAD_RENDER |
                        FT_LOAD_TARGET_MONO )

    bitmap = face.glyph.bitmap
    width  = face.glyph.bitmap.width
    rows   = face.glyph.bitmap.rows
    pitch  = face.glyph.bitmap.pitch

    data = []
    for i in range(bitmap.rows):
        row = []
        for j in range(bitmap.pitch):
            row.extend(bits(bitmap.buffer[i*bitmap.pitch+j]))
        data.extend(row[:bitmap.width])
    Z = numpy.array(data).reshape(bitmap.rows, bitmap.width)
    plt.imshow(Z, interpolation='nearest', cmap=plt.cm.gray, origin='lower')
    plt.show()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import freetype
import numpy as np
from PIL import Image

def convert_bgra_to_rgb(buf):
    blue = buf[:,:,0]
    green = buf[:,:,1]
    red = buf[:,:,2]
    return np.dstack((red, green, blue))

face = freetype.Face("/System/Library/Fonts/Apple Color Emoji.ttc")
face.set_char_size( 20*64 )
face.load_char('😀', freetype.FT_LOAD_COLOR)
bitmap = face.glyph.bitmap
bitmap = np.array(bitmap.buffer, dtype=np.uint8).reshape((bitmap.rows,bitmap.width,4))
rgb = convert_bgra_to_rgb(bitmap)
im = Image.fromarray(rgb)
im.show()



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
#
#  FreeType high-level python API - Copyright 2011-2015 Nicolas P. Rougier
#  Distributed under the terms of the new BSD license.
#
# -----------------------------------------------------------------------------

"""
A list of values to identify various types of LCD filters.


FT_LCD_FILTER_NONE

  Do not perform filtering. When used with subpixel rendering, this results in
  sometimes severe color fringes.


FT_LCD_FILTER_DEFAULT

  The default filter reduces color fringes considerably, at the cost of a
  slight blurriness in the output.


FT_LCD_FILTER_LIGHT

  The light filter is a variant that produces less blurriness at the cost of
  slightly more color fringes than the default one. It might be better,
  depending on taste, your monitor, or your personal vision.


FT_LCD_FILTER_LEGACY

  This filter corresponds to the original libXft color filter. It provides high
  contrast output but can exhibit really bad color fringes if glyphs are not
  extremely well hinted to the pixel grid. In other words, it only works well
  if the TrueType bytecode interpreter is enabled and high-quality hinted fonts
  are used.

  This filter is only provided for comparison purposes, and might be disabled
  or stay unsupported in the future.
"""

FT_LCD_FILTERS = {'FT_LCD_FILTER_NONE'    : 0,
                  'FT_LCD_FILTER_DEFAULT' : 1,
                  'FT_LCD_FILTER_LIGHT'   : 2,
                  'FT_LCD_FILTER_LEGACY'  : 16}
globals().update(FT_LCD_FILTERS)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  OT-SVG example
#
#  Copyright 2023 Hin-Tak Leung
#  Distributed under the terms of the new BSD license.

# This is largely a python re-write of freetype2-demos:src/rsvg-port.c .
# It is designed to be embeddable as a module in another python script .
#
#     To use, rename this to, for example "otsvg.py", then insert
#
#     '''
#         from otsvg import hooks
#
#         library = get_handle()
#         FT_Property_Set( library, b"ot-svg", b"svg-hooks", byref(hooks)
#     '''
#
#     in your script, and add "FT_LOAD_COLOR | FT_LOAD_RENDER"
#     to load_glyph() or load_char() calls. As in "__main__" below,
#     but also in the longer example of hb-view-ot-svg.py
#     in https://github.com/HinTak/harfbuzz-python-demos/ .
#
# Limitation: it is necessary to have "_state" as a module-level global
# partially (in svg_init/svg_free, not in svg_render/svg_preset_slot)
# to stop python destroying it when execution is in the c-side.
#
# Note:
#    Strictly-speaking, cairo.FORMAT_ARGB32 is host-order,
#    while freetype.FT_PIXEL_MODE_BGRA is small-endian. They are
#    different on big-endian platforms. The below
#    works in all circumstances, only because the bitmap is both
#    generated by cairo at the beginning and also consumed
#    by cairo in the end.
#
#    The use of "pythonapi.PyMemoryView_FromMemory" is flaky -
#    specific to CPython 3.3+.
#
#    Uses librsvg 2.52+ only. Or deprecation warnings per run.
#
#    Not using Rsvg.Handle.render_document() .
#    Rsvg.Handle.render_layer(..., None, ...) does the same work.
#
#    The original C-code does not check
#     start_glyph_id <= glyph_index <= end_glyph_id.value .

import gi
gi.require_version('Rsvg', '2.0')

from gi.repository import Rsvg as rsvg

from freetype import *
from cairo import * # cairo.Matrix shadows freetype.Matrix

from math import ceil

_state = None

def svg_init(ctx):
    global _state
    _state = {}
    ctx.contents.value = _state
    return FT_Err_Ok

def svg_free(ctx):
    global _state
    _state = None
    # "None" is strictly speaking a special pyobject,
    # this line does not do what it should, i.e. setting the
    # pointer to NULL.
    ctx.contents = None
    return # void

def svg_render(slot, ctx):
    state = ctx.contents.value
    #pythonapi is imported from ctypes
    pythonapi.PyMemoryView_FromMemory.argtypes = (c_char_p, c_ssize_t, c_int)
    pythonapi.PyMemoryView_FromMemory.restype = py_object
    surface = ImageSurface.create_for_data( pythonapi.PyMemoryView_FromMemory(cast(slot.contents.bitmap.buffer, c_char_p),
                                                                              slot.contents.bitmap.rows * slot.contents.bitmap.pitch, 0x200),
                                            FORMAT_ARGB32,
                                            slot.contents.bitmap.width,
                                            slot.contents.bitmap.rows,
                                            slot.contents.bitmap.pitch )
    cr     = Context( surface )
    cr.translate( -state['x'], -state['y'] )

    cr.set_source_surface( state['rec_surface'] ) # 0,0 is default
    cr.paint()

    surface.flush()

    slot.contents.bitmap.pixel_mode = FT_PIXEL_MODE_BGRA
    slot.contents.bitmap.num_grays  = 256
    slot.contents.format            = FT_GLYPH_FORMAT_BITMAP

    state['rec_surface'] = None # Let python destroy the surface

    return FT_Err_Ok

def svg_preset_slot(slot, cached, ctx):
    state = ctx.contents.value

    document = ctypes.cast(slot.contents.other, FT_SVG_Document)

    metrics        = SizeMetrics(document.contents.metrics)

    units_per_EM   = FT_UShort(document.contents.units_per_EM)
    end_glyph_id   = FT_UShort(document.contents.end_glyph_id)
    start_glyph_id = FT_UShort(document.contents.start_glyph_id)

    dimension_svg = rsvg.DimensionData()

    handle = rsvg.Handle.new_from_data( ctypes.string_at(document.contents.svg_document, # not terminated
                                                         size=document.contents.svg_document_length)
                                       )

    (out_has_width, out_width,
     out_has_height, out_height,
     out_has_viewbox, out_viewbox) = handle.get_intrinsic_dimensions()

    if ( out_has_viewbox == True ):
        dimension_svg.width  = out_viewbox.width
        dimension_svg.height = out_viewbox.height
    else:
        # "out_has_width" and "out_has_height" are True always
        dimension_svg.width  = units_per_EM.value
        dimension_svg.height = units_per_EM.value

        if (( out_width.length  != 1) or (out_height.length != 1 )):
            dimension_svg.width  = out_width.length
            dimension_svg.height = out_height.length

    x_svg_to_out = metrics.x_ppem / dimension_svg.width
    y_svg_to_out = metrics.y_ppem / dimension_svg.height

    state['rec_surface'] = RecordingSurface( Content.COLOR_ALPHA, None )

    rec_cr = Context( state['rec_surface'] )

    xx =  document.contents.transform.xx / ( 1 << 16 )
    xy = -document.contents.transform.xy / ( 1 << 16 )
    yx = -document.contents.transform.yx / ( 1 << 16 )
    yy =  document.contents.transform.yy / ( 1 << 16 )

    x0 =  document.contents.delta.x / 64 * dimension_svg.width / metrics.x_ppem
    y0 = -document.contents.delta.y / 64 * dimension_svg.height / metrics.y_ppem

    transform_matrix = Matrix(xx, yx, xy, yy, x0, y0) # cairo.Matrix

    rec_cr.scale( x_svg_to_out, y_svg_to_out )

    rec_cr.transform( transform_matrix )

    viewport = rsvg.Rectangle()
    viewport.x = 0
    viewport.y = 0
    viewport.width = dimension_svg.width
    viewport.height = dimension_svg.height

    str = None # render whole document - not using Handle.render_document()
    if ( start_glyph_id.value < end_glyph_id.value ):
        str = "#glyph%u" % (slot.contents.glyph_index )

    handle.render_layer( rec_cr, str, viewport )

    (state['x'], state['y'], width, height) = state['rec_surface'].ink_extents()

    slot.contents.bitmap_left = int(state['x'])
    slot.contents.bitmap_top  = int(-state['y'])

    slot.contents.bitmap.rows  = ceil( height )
    slot.contents.bitmap.width = ceil( width )

    slot.contents.bitmap.pitch = slot.contents.bitmap.width * 4

    slot.contents.bitmap.pixel_mode = FT_PIXEL_MODE_BGRA

    metrics_width  = width
    metrics_height = height

    horiBearingX =  state['x']
    horiBearingY = -state['y']

    vertBearingX = slot.contents.metrics.horiBearingX / 64.0 - slot.contents.metrics.horiAdvance / 64.0 / 2
    vertBearingY = ( slot.contents.metrics.vertAdvance / 64.0 - slot.contents.metrics.height / 64.0 ) / 2

    slot.contents.metrics.width  = int(round( metrics_width * 64 ))
    slot.contents.metrics.height = int(round( metrics_height * 64 ))

    slot.contents.metrics.horiBearingX = int( horiBearingX * 64 )
    slot.contents.metrics.horiBearingY = int( horiBearingY * 64 )
    slot.contents.metrics.vertBearingX = int( vertBearingX * 64 )
    slot.contents.metrics.vertBearingY = int( vertBearingY * 64 )

    if ( slot.contents.metrics.vertAdvance == 0 ):
        slot.contents.metrics.vertAdvance = int( metrics_height * 1.2 * 64 )

    if ( cached == False ):
        state['rec_surface'] = None
        state['x'] = 0
        state['y'] = 0

    return FT_Err_Ok

hooks = SVG_RendererHooks(svg_init=SVG_Lib_Init_Func(svg_init),
                          svg_free=SVG_Lib_Free_Func(svg_free),
                          svg_render=SVG_Lib_Render_Func(svg_render),
                          svg_preset_slot=SVG_Lib_Preset_Slot_Func(svg_preset_slot))

if __name__ == '__main__':
    import sys
    execname = sys.argv[0]

    if len(sys.argv) < 2:
        print("Example usage: %s TrajanColor-Concept.otf" % execname)
        exit(1)

    face = Face(sys.argv[1])

    face.set_char_size( 160*64 )
    library = get_handle()
    FT_Property_Set( library, b"ot-svg", b"svg-hooks", byref(hooks) ) # python 3 only syntax
    face.load_char('A', FT_LOAD_COLOR | FT_LOAD_RENDER )

    bitmap = face.glyph.bitmap
    width = face.glyph.bitmap.width
    rows = face.glyph.bitmap.rows

    if ( face.glyph.bitmap.pitch != width * 4 ):
        raise RuntimeError('pitch != width * 4 for color bitmap: Please report this.')

    I = ImageSurface.create_for_data( pythonapi.PyMemoryView_FromMemory(cast(bitmap._FT_Bitmap.buffer, c_char_p),
                                                                        bitmap.rows * bitmap.pitch,
                                                                        0x200), # Read-Write
                                      FORMAT_ARGB32,
                                      width, rows,
                                      bitmap.pitch )
    surface = ImageSurface(FORMAT_ARGB32, 2*width, rows)
    ctx = Context(surface)

    ctx.set_source_surface(I, 0, 0)
    ctx.paint()

    ctx.set_source_surface(I, width/2, 0)
    ctx.paint()

    ctx.set_source_surface(I, width , 0)
    ctx.paint()

    surface.write_to_png("ot-svg-example.png")

    from PIL import Image
    Image.open("ot-svg-example.png").show()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
#
#  FreeType high-level python API - Copyright 2011 Nicolas P. Rougier
#  Distributed under the terms of the new BSD license.
#
#  - The code is incomplete and over-simplified, as it ignores the 3rd order
#    bezier curve bit and always intepolate between off-curve points.
#    This is only correct for truetype fonts (which only use 2nd order bezier curves).
#  - Also it seems to assume the first point is always on curve; this is
#    unusual but legal.
#
# -----------------------------------------------------------------------------
'''
Show how to access glyph outline description.
'''
from freetype import *

if __name__ == '__main__':
    import numpy
    import matplotlib.pyplot as plt
    from matplotlib.path import Path
    import matplotlib.patches as patches

    face = Face('./Vera.ttf')
    face.set_char_size( 48*64 )
    face.load_char('S')
    slot = face.glyph

    outline = slot.outline
    points = numpy.array(outline.points, dtype=[('x',float), ('y',float)])
    x, y = points['x'], points['y']

    figure = plt.figure(figsize=(8,10))
    axis = figure.add_subplot(111)
    #axis.scatter(points['x'], points['y'], alpha=.25)
    start, end = 0, 0

    VERTS, CODES = [], []
    # Iterate over each contour
    for i in range(len(outline.contours)):
        end    = outline.contours[i]
        points = outline.points[start:end+1]
        points.append(points[0])
        tags   = outline.tags[start:end+1]
        tags.append(tags[0])

        segments = [ [points[0],], ]
        for j in range(1, len(points) ):
            segments[-1].append(points[j])
            if tags[j] & (1 << 0) and j < (len(points)-1):
                segments.append( [points[j],] )
        verts = [points[0], ]
        codes = [Path.MOVETO,]
        for segment in segments:
            if len(segment) == 2:
                verts.extend(segment[1:])
                codes.extend([Path.LINETO])
            elif len(segment) == 3:
                verts.extend(segment[1:])
                codes.extend([Path.CURVE3, Path.CURVE3])
            else:
                verts.append(segment[1])
                codes.append(Path.CURVE3)
                for i in range(1,len(segment)-2):
                    A,B = segment[i], segment[i+1]
                    C = ((A[0]+B[0])/2.0, (A[1]+B[1])/2.0)
                    verts.extend([ C, B ])
                    codes.extend([ Path.CURVE3, Path.CURVE3])
                verts.append(segment[-1])
                codes.append(Path.CURVE3)
        VERTS.extend(verts)
        CODES.extend(codes)
        start = end+1


    # Draw glyph lines
    path = Path(VERTS, CODES)
    glyph = patches.PathPatch(path, facecolor='.75', lw=1)

    # Draw "control" lines
    for i, code in enumerate(CODES):
        if code == Path.CURVE3:
            CODES[i] = Path.LINETO
    path = Path(VERTS, CODES)
    patch = patches.PathPatch(path, ec='.5', fill=False, ls='dashed', lw=1 )

    axis.add_patch(patch)
    axis.add_patch(glyph)

    axis.set_xlim(x.min()-100, x.max()+100)
    plt.xticks([])
    axis.set_ylim(y.min()-100, y.max()+100)
    plt.yticks([])
    plt.show()



# This will build a FreeType binary for bundling with freetype-py on Windows,
# Linux and macOS. The environment determines if it's a 32 or 64 bit build (on
# Windows, set $env:PYTHON_ARCH="64" for a x64 build; on Linux, set it to "32"
# to explicitly make a 32 bit build).

# It can be used stand-alone, in conjunction with setup.py and on CI services
# like Travis and Appveyor. You need CMake and some C/C++ compiler suite (e.g.
# GCC, Visual Studio Community 2017, ...)

import distutils.dir_util
import distutils.file_util
import distutils.spawn
import glob
import hashlib
import os
import shutil
import ssl
import subprocess
import sys
import tarfile
import urllib.request
from os import path
import platform

# Needed for the GitHub Actions macOS CI runner, which appears to come without CAs.
import certifi

FREETYPE_HOST = "https://mirrors.sarata.com/non-gnu/freetype/"
FREETYPE_TARBALL = "freetype-2.13.2.tar.xz"
FREETYPE_URL = FREETYPE_HOST + FREETYPE_TARBALL
FREETYPE_SHA256 = "12991c4e55c506dd7f9b765933e62fd2be2e06d421505d7950a132e4f1bb484d"
HARFBUZZ_HOST = "https://github.com/harfbuzz/harfbuzz/releases/download/8.3.0/"
HARFBUZZ_TARBALL = "harfbuzz-8.3.0.tar.xz"
HARFBUZZ_URL = HARFBUZZ_HOST + HARFBUZZ_TARBALL
HARFBUZZ_SHA256 = "109501eaeb8bde3eadb25fab4164e993fbace29c3d775bcaa1c1e58e2f15f847"

root_dir = "."
build_dir = path.join(root_dir, "build")
# CMake requires an absolute path to a prefix.
prefix_dir = path.abspath(path.join(build_dir, "local"))
lib_dir = path.join(prefix_dir, "lib")
build_dir_ft = path.join(build_dir, FREETYPE_TARBALL.split(".tar")[0], "build")
build_dir_hb = path.join(build_dir, HARFBUZZ_TARBALL.split(".tar")[0], "build")

CMAKE_GLOBAL_SWITCHES = (
    "-DCMAKE_COLOR_MAKEFILE=false "
    '-DCMAKE_PREFIX_PATH="{}" '
    '-DCMAKE_INSTALL_PREFIX="{}" '
).format(prefix_dir, prefix_dir)

# Try to use Ninja to build things if it's available. Much faster.
# On Windows, I first need to figure out how to make it aware of VC, bitness,
# etc.
if sys.platform != "win32" and distutils.spawn.find_executable("ninja"):
    CMAKE_GLOBAL_SWITCHES += "-G Ninja "

bitness = None

if sys.platform == "win32":
    if os.environ.get("PYTHON_ARCH", "") == "64":
        print("# Making a 64 bit build.")
        bitness = 64
        CMAKE_GLOBAL_SWITCHES += (
            "-DCMAKE_GENERATOR_PLATFORM=x64 " '-DCMAKE_GENERATOR_TOOLSET="host=x64" '
        )
    else:
        print("# Making a 32 bit build.")
        bitness = 32

if sys.platform == "darwin":
    print("# Making a 64 bit build.")
    CMAKE_GLOBAL_SWITCHES += (
        '-DCMAKE_OSX_ARCHITECTURES="x86_64;arm64" '
        '-DCMAKE_OSX_DEPLOYMENT_TARGET="10.9" '
        '-DCMAKE_C_FLAGS="-O2" '
        '-DCMAKE_CXX_FLAGS="-O2" '
    )
    bitness = 64

if "linux" in sys.platform:
    c_flags = cxx_flags = "-O2"
    ld_flags = ""
    if platform.machine() == "x86_64":
        if (
            os.environ.get("PYTHON_ARCH", "") == "32"
            or platform.architecture() == "32bit"
        ):
            print("# Making a 32 bit build.")
            # On a 64 bit Debian/Ubuntu, this needs gcc-multilib and g++-multilib.
            # On a 64 bit Fedora, install glibc-devel and libstdc++-devel.
            c_flags = "-m32 {}".format(c_flags)
            cxx_flags = "-m32 {}".format(cxx_flags)
            ld_flags = "-m32"
            bitness = 32
        else:
            print("# Making a 64 bit build.")
            c_flags = "-m64 {}".format(c_flags)
            cxx_flags = "-m64 {}".format(cxx_flags)
            ld_flags = "-m64"
            bitness = 64
    else:
        print("# Making a '{}' build.".format(platform.machine()))
        if platform.architecture() == "32bit":
            bitness = 32
        else:
            bitness = 64
    CMAKE_GLOBAL_SWITCHES += (
        '-DCMAKE_C_FLAGS="{}" '.format(c_flags) +
        '-DCMAKE_CXX_FLAGS="{}" '.format(cxx_flags) +
        '-DCMAKE_LD_FLAGS="{}" '.format(ld_flags)
    )


def shell(cmd, cwd=None):
    """Run a shell command specified by cmd string."""
    subprocess.run(cmd, shell=True, check=True, cwd=cwd)


def download(url, target_path):
    """Download url to target_path."""
    print("Downloading {}...".format(url))
    # Create a custom context and fill in certifi CAs because GitHub Action's macOS CI
    # runners don't seem to have certificates installed, leading to a download
    # failure...
    context = ssl.create_default_context(
        ssl.Purpose.SERVER_AUTH, cafile=certifi.where()
    )
    with urllib.request.urlopen(url, context=context) as response, open(
        target_path, "wb"
    ) as f:
        shutil.copyfileobj(response, f)


def ensure_downloaded(url, sha256_sum):
    """Download .tar.bz2 tarball at url unless already present, extract."""
    filename = path.basename(url)
    tarball = path.join(build_dir, filename)
    tarball_dir = filename.split(".tar")[0]

    if not path.exists(tarball):
        download(url, tarball)

    if not path.exists(path.join(build_dir, tarball_dir, "CMakeLists.txt")):
        hasher = hashlib.sha256()
        with open(tarball, "rb") as tb:
            hasher.update(tb.read())
        assert hasher.hexdigest() == sha256_sum

        with tarfile.open(tarball, "r:xz") as tb:
            tb.extractall(build_dir)


distutils.dir_util.mkpath(build_dir)
distutils.dir_util.mkpath(prefix_dir)
distutils.dir_util.mkpath(build_dir_ft)
distutils.dir_util.mkpath(build_dir_hb)

ensure_downloaded(FREETYPE_URL, FREETYPE_SHA256)
ensure_downloaded(HARFBUZZ_URL, HARFBUZZ_SHA256)

print("# First, build FreeType without Harfbuzz support")
shell(
    "cmake -DBUILD_SHARED_LIBS=OFF "
    "-DCMAKE_DISABLE_FIND_PACKAGE_HarfBuzz=TRUE -DFT_WITH_HARFBUZZ=OFF "
    "-DCMAKE_DISABLE_FIND_PACKAGE_PNG=TRUE "
    "-DCMAKE_DISABLE_FIND_PACKAGE_BZip2=TRUE "
    "-DCMAKE_DISABLE_FIND_PACKAGE_ZLIB=TRUE "
    "-DCMAKE_DISABLE_FIND_PACKAGE_BrotliDec=TRUE "
    "{} ..".format(CMAKE_GLOBAL_SWITCHES),
    cwd=build_dir_ft,
)
shell("cmake --build . --config Release --target install --parallel", cwd=build_dir_ft)

print("\n# Next, build Harfbuzz and point it to the FreeType we just build.")
shell(
    "cmake -DBUILD_SHARED_LIBS=OFF " +
    # https://stackoverflow.com/questions/3961446
    ("-DCMAKE_POSITION_INDEPENDENT_CODE=ON " if bitness > 32 else "")
    + "-DHB_HAVE_FREETYPE=ON -DHB_HAVE_GLIB=OFF -DHB_HAVE_CORETEXT=OFF "
    "{} ..".format(CMAKE_GLOBAL_SWITCHES),
    cwd=build_dir_hb,
)
shell("cmake --build . --config Release --target install --parallel", cwd=build_dir_hb)

print("\n# Lastly, rebuild FreeType, this time with Harfbuzz support.")
harfbuzz_includes = path.join(prefix_dir, "include", "harfbuzz")
shell(
    "cmake -DBUILD_SHARED_LIBS=ON "
    "-DCMAKE_DISABLE_FIND_PACKAGE_HarfBuzz=FALSE -DFT_WITH_HARFBUZZ=ON "
    "-DCMAKE_DISABLE_FIND_PACKAGE_PNG=TRUE "
    "-DCMAKE_DISABLE_FIND_PACKAGE_BZip2=TRUE "
    "-DCMAKE_DISABLE_FIND_PACKAGE_ZLIB=TRUE "
    "-DCMAKE_DISABLE_FIND_PACKAGE_BrotliDec=TRUE "
    '-DPKG_CONFIG_EXECUTABLE="" '  # Prevent finding system libraries
    '-DHARFBUZZ_INCLUDE_DIRS="{}" '
    "-DSKIP_INSTALL_HEADERS=ON "
    "{} ..".format(harfbuzz_includes, CMAKE_GLOBAL_SWITCHES),
    cwd=build_dir_ft,
)
shell("cmake --build . --config Release --target install --parallel", cwd=build_dir_ft)

# Move libraries from PREFIX/bin to PREFIX/lib if need be (Windows DLLs are
# treated as runtimes and may end up in bin/). This keeps setup.py simple.
target_dir = path.join(prefix_dir, "lib")
bin_so = glob.glob(path.join(prefix_dir, "bin", "*freetype*"))
bin_so.extend(glob.glob(path.join(prefix_dir, "lib64", "*freetype*")))
bin_so.extend(glob.glob(path.join(prefix_dir, "lib", "freetype*.dll")))
for so in bin_so:
    so_target_name = path.basename(so)
    if not so_target_name.startswith("lib"):
        so_target_name = "lib" + so_target_name
    lib_path = path.join(target_dir, so_target_name)

    print("# Moving '{}' to '{}'".format(so, lib_path))
    if not path.exists(target_dir):
        os.makedirs(target_dir)
    os.replace(so, lib_path)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# shorter Unicode Variation Sequences example from #195, with comparison to hb-shape.

# The test file used and known to work is v2.004 of https://github.com/adobe-fonts/source-han-sans/blob/release/OTF/Japanese/SourceHanSans-Regular.otf

import importlib
uvs = importlib.import_module("unicode-variation-sequences")

read_freetype_cmap = uvs.read_freetype_cmap

if __name__ == "__main__":
    import os, io
    directory   = os.path.dirname(__file__)
    font_path   = os.path.join(directory, 'SourceHanSans-Regular.otf')
    memory_file = io.BytesIO()
    with open(font_path, 'rb') as fontfile:
        memory_file.write(fontfile.read())
        memory_file.seek(0)
    
    from freetype import *
    library = FT_Library()
    error   = FT_Init_FreeType(byref(library))
    if error: raise FT_Exception(error)

    freetype_face = FT_Face()
    data          = memory_file.getvalue()
    error         = FT_New_Memory_Face(library, data, len(data), 0, byref(freetype_face))
    if error: raise FT_Exception(error)

    all_freetype_characters  = read_freetype_cmap(freetype_face)
    # {'邉' : ['邉󠄁', '邉󠄂', '邉󠄃', '邉󠄄', '邉󠄅', '邉󠄆', '邉󠄇', '邉󠄈', '邉󠄉', '邉󠄊', '邉󠄋', '邉󠄌', '邉󠄍', '邉󠄎', '邉󠄀']}
    
    print(all_freetype_characters['邉'], all_freetype_characters['邉󠄁'], all_freetype_characters['邉󠄂'], all_freetype_characters['邉󠄃'], 
          all_freetype_characters['邉󠄄'], all_freetype_characters['邉󠄅'], all_freetype_characters['邉󠄆'], all_freetype_characters['邉󠄇'],
          all_freetype_characters['邉󠄈'], all_freetype_characters['邉󠄉'], all_freetype_characters['邉󠄊'], all_freetype_characters['邉󠄋'],
          all_freetype_characters['邉󠄌'], all_freetype_characters['邉󠄍'], all_freetype_characters['邉󠄎'], all_freetype_characters['邉󠄀'],
          sep='|')
    print('The above should be identical to the output of this hb-shape command:')
    print('    hb-shape --no-glyph-names --no-positions --no-clusters --no-advances SourceHanSans-Regular.otf "邉邉󠄁邉󠄂邉󠄃邉󠄄邉󠄅邉󠄆邉󠄇邉󠄈邉󠄉邉󠄊邉󠄋邉󠄌邉󠄍邉󠄎邉󠄀"')



# skia_glfw_module
# This file created by 2023 Hin-Tak Leung; but Copyright skia-python project:

# Adapted from https://kyamagu.github.io/skia-python/tutorial/canvas.html#opengl-window

# Typical usage:
#
#     from skia_glfw_module import glfw_window, skia_surface
#     ....
#     with glfw_window(WIDTH, HEIGHT) as window:
#     ...
#          with skia_surface(window) as surface:
#              with surface as canvas:
#                  canvas.drawStuff()
#
#              surface.flushAndSubmit()
#              glfw.swap_buffers(window)
#

import contextlib, glfw
import skia
from OpenGL import GL

@contextlib.contextmanager
def glfw_window(WIDTH, HEIGHT):
    if not glfw.init():
        raise RuntimeError('glfw.init() failed')
    glfw.window_hint(glfw.STENCIL_BITS, 8)
    window = glfw.create_window(WIDTH, HEIGHT, '', None, None)
    glfw.make_context_current(window)
    yield window
    glfw.terminate()

@contextlib.contextmanager
def skia_surface(window):
    context = skia.GrDirectContext.MakeGL()
    (fb_width, fb_height) = glfw.get_framebuffer_size(window)
    backend_render_target = skia.GrBackendRenderTarget(
        fb_width,
        fb_height,
        0,  # sampleCnt
        0,  # stencilBits
        skia.GrGLFramebufferInfo(0, GL.GL_RGBA8))
    surface = skia.Surface.MakeFromBackendRenderTarget(
        context, backend_render_target, skia.kBottomLeft_GrSurfaceOrigin,
        skia.kRGBA_8888_ColorType, skia.ColorSpace.MakeSRGB())
    assert surface is not None
    yield surface
    context.abandonContext()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
#
#  FreeType high-level python API - Copyright 2011-2015 Nicolas P. Rougier
#  Distributed under the terms of the new BSD license.
#
# -----------------------------------------------------------------------------
import math
import numpy as np
from freetype import *
import matplotlib.pyplot as plt


def make_label(text, filename, size=12, angle=0):
    '''
    Parameters:
    -----------
    text : string
        Text to be displayed
    filename : string
        Path to a font
    size : int
        Font size in 1/64th points
    angle : float
        Text angle in degrees
    '''
    face = Face(filename)
    face.set_char_size( size*64 )
    angle = (angle/180.0)*math.pi
    matrix  = FT_Matrix( (int)( math.cos( angle ) * 0x10000 ),
                         (int)(-math.sin( angle ) * 0x10000 ),
                         (int)( math.sin( angle ) * 0x10000 ),
                         (int)( math.cos( angle ) * 0x10000 ))
    flags = FT_LOAD_RENDER
    pen = FT_Vector(0,0)
    FT_Set_Transform( face._FT_Face, byref(matrix), byref(pen) )
    previous = 0
    xmin, xmax = 0, 0
    ymin, ymax = 0, 0
    for c in text:
        face.load_char(c, flags)
        kerning = face.get_kerning(previous, c)
        previous = c
        bitmap = face.glyph.bitmap
        pitch  = face.glyph.bitmap.pitch
        width  = face.glyph.bitmap.width
        rows   = face.glyph.bitmap.rows
        top    = face.glyph.bitmap_top
        left   = face.glyph.bitmap_left
        pen.x += kerning.x
        x0 = (pen.x >> 6) + left
        x1 = x0 + width
        y0 = (pen.y >> 6) - (rows - top)
        y1 = y0 + rows
        xmin, xmax = min(xmin, x0),  max(xmax, x1)
        ymin, ymax = min(ymin, y0), max(ymax, y1)
        pen.x += face.glyph.advance.x
        pen.y += face.glyph.advance.y

    L = np.zeros((ymax-ymin, xmax-xmin),dtype=np.ubyte)
    previous = 0
    pen.x, pen.y = 0, 0
    for c in text:
        face.load_char(c, flags)
        kerning = face.get_kerning(previous, c)
        previous = c
        bitmap = face.glyph.bitmap
        pitch  = face.glyph.bitmap.pitch
        width  = face.glyph.bitmap.width
        rows   = face.glyph.bitmap.rows
        top    = face.glyph.bitmap_top
        left   = face.glyph.bitmap_left
        pen.x += kerning.x
        x = (pen.x >> 6) - xmin + left
        y = (pen.y >> 6) - ymin - (rows - top)
        data = []
        for j in range(rows):
            data.extend(bitmap.buffer[j*pitch:j*pitch+width])
        if len(data):
            Z = np.array(data,dtype=np.ubyte).reshape(rows, width)
            L[y:y+rows,x:x+width] |= Z[::-1,::1]
        pen.x += face.glyph.advance.x
        pen.y += face.glyph.advance.y

    return L


if __name__ == '__main__':
    from PIL import Image

    n_words = 100
    H, W, dpi = 600, 800, 72.0
    I = np.zeros((H, W, 3), dtype=np.ubyte)
    S = np.random.normal(0,1,n_words)
    S = (S-S.min())/(S.max()-S.min())
    S = np.sort(1-np.sqrt(S))[::-1]
    sizes = (12 + S*48).astype(int).tolist()

    def spiral():
        eccentricity = 1.5
        radius = 8
        step = 0.1
        t = 0
        while True:
            t += step
            yield eccentricity*radius*t*math.cos(t), radius*t*math.sin(t)

    fails = 0
    for size in sizes:
        angle = np.random.randint(-25,25)
        L = make_label('Hello', './Vera.ttf', size, angle=angle)
        h,w = L.shape
        if h < H and w < W:
            x0 = W//2 + (np.random.uniform()-.1)*50
            y0 = H//2 + (np.random.uniform()-.1)*50
            for dx,dy in spiral():
                c = .25+.75*np.random.random()
                x = int(x0+dx)
                y = int(y0+dy)
                if x <= w//2 or y <= h//2 or x >= (W-w//2) or y >= (H-h//2):
                    fails += 1
                    break
                if (I[y-h//2:y-h//2+h, x-w//2:x-w//2+w,0] * L).sum() == 0:
                    I[y-h//2:y-h//2+h, x-w//2:x-w//2+w,0] |= (c * L).astype('ubyte')
                    I[y-h//2:y-h//2+h, x-w//2:x-w//2+w,1] |= (c * L).astype('ubyte')
                    I[y-h//2:y-h//2+h, x-w//2:x-w//2+w,2] |= (c * L).astype('ubyte')
                    break

    print ("Number of fails: {}".format(fails))
    fig = plt.figure(figsize=(W/dpi,H/dpi), dpi=dpi)
    ax = fig.add_axes([0,0,1,1], frameon=False)
    ax.imshow(I, interpolation='nearest', cmap=plt.cm.gray, origin='upper')
    #plt.axis('off')
    plt.show()
    I = Image.fromarray(I[::-1,::1,::1], mode='RGB')
    I.save('wordle.png')



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
#
#  FreeType high-level python API - Copyright 2011-2015 Nicolas P. Rougier
#  Distributed under the terms of the new BSD license.
#
# -----------------------------------------------------------------------------
from __future__ import print_function
from __future__ import division
from freetype import *

def platform_name(platform_id):
    for key, value in TT_PLATFORMS.items():
        if value == platform_id:
            return key
    return 'Unknown platform'

def encoding_name(platform_id, encoding_id):
    if platform_id == TT_PLATFORM_APPLE_UNICODE:
        encodings = TT_APPLE_IDS
    elif platform_id == TT_PLATFORM_MACINTOSH:
        encodings = TT_MAC_IDS
    elif platform_id == TT_PLATFORM_MICROSOFT:
        encodings = TT_MS_IDS
    elif platform_id == TT_PLATFORM_ADOBE:
        encodings = TT_ADOBE_IDS
    else:
        return 'Unknown encoding'
    for key, value in encodings.items():
        if value == encoding_id:
            return key
    return 'Unknown encoding'

def language_name(platform_id, language_id):
    if platform_id == TT_PLATFORM_MACINTOSH:
        languages = TT_MAC_LANGIDS
    elif platform_id == TT_PLATFORM_MICROSOFT:
        languages = TT_MS_LANGIDS
    else:
        return 'Unknown language'
    for key, value in languages.items():
        if value == language_id:
            return key
    return 'Unknown language'


if __name__ == '__main__':
    import os, sys

    if len(sys.argv) < 2:
        print("Usage: %s font_filename" % sys.argv[0])
        sys.exit()
    face = Face(sys.argv[1])

    name = face.get_sfnt_name(0)
    print( 'platform_id:', platform_name(name.platform_id) )
    print( 'encoding_id:', encoding_name(name.platform_id,
                                        name.encoding_id) )
    print( 'language_id:', language_name(name.platform_id,
                                        name.language_id) )
    for i in range(face.sfnt_name_count):
        name = face.get_sfnt_name(i).string
        print(i, name.decode('utf-8', 'ignore'))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
#
#  FreeType high-level python API - Copyright 2011-2015 Nicolas P. Rougier
#  Distributed under the terms of the new BSD license.
#
# -----------------------------------------------------------------------------
import numpy
from freetype import *
import OpenGL.GL as gl
import OpenGL.GLUT as glut

base, texid = 0, 0
text  = '''Hello World !'''

def on_display( ):
    global texid
    gl.glClearColor(1,1,1,1)
    gl.glClear(gl.GL_COLOR_BUFFER_BIT | gl.GL_DEPTH_BUFFER_BIT)
    gl.glBindTexture( gl.GL_TEXTURE_2D, texid )
    gl.glColor(0,0,0,1)
    gl.glPushMatrix( )
    gl.glTranslate( 10, 100, 0 )
    gl.glPushMatrix( )
    gl.glListBase( base+1 )
    gl.glCallLists( [ord(c) for c in text] )
    gl.glPopMatrix( )
    gl.glPopMatrix( )
    glut.glutSwapBuffers( )

def on_reshape( width, height ):
    gl.glViewport( 0, 0, width, height )
    gl.glMatrixMode( gl.GL_PROJECTION )
    gl.glLoadIdentity( )
    gl.glOrtho( 0, width, 0, height, -1, 1 )
    gl.glMatrixMode( gl.GL_MODELVIEW )
    gl.glLoadIdentity( )

def on_keyboard( key, x, y ):
    if key == '\033': sys.exit( )

def makefont(filename, size):
    global texid

    # Load font  and check it is monotype
    face = Face(filename)
    face.set_char_size( size*64 )
    if not face.is_fixed_width:
        raise 'Font is not monotype'

    # Determine largest glyph size
    width, height, ascender, descender = 0, 0, 0, 0
    for c in range(32,128):
        face.load_char( chr(c), FT_LOAD_RENDER | FT_LOAD_FORCE_AUTOHINT )
        bitmap    = face.glyph.bitmap
        width     = max( width, bitmap.width )
        ascender  = max( ascender, face.glyph.bitmap_top )
        descender = max( descender, bitmap.rows-face.glyph.bitmap_top )
    height = ascender+descender

    # Generate texture data
    Z = numpy.zeros((height*6, width*16), dtype=numpy.ubyte)
    for j in range(6):
        for i in range(16):
            face.load_char(chr(32+j*16+i), FT_LOAD_RENDER | FT_LOAD_FORCE_AUTOHINT )
            bitmap = face.glyph.bitmap
            x = i*width  + face.glyph.bitmap_left
            y = j*height + ascender - face.glyph.bitmap_top
            Z[y:y+bitmap.rows,x:x+bitmap.width].flat = bitmap.buffer

    # Bound texture
    texid = gl.glGenTextures(1)
    gl.glBindTexture( gl.GL_TEXTURE_2D, texid )
    gl.glTexParameterf( gl.GL_TEXTURE_2D, gl.GL_TEXTURE_MAG_FILTER, gl.GL_LINEAR )
    gl.glTexParameterf( gl.GL_TEXTURE_2D, gl.GL_TEXTURE_MIN_FILTER, gl.GL_LINEAR )
    gl.glTexImage2D( gl.GL_TEXTURE_2D, 0, gl.GL_ALPHA, Z.shape[1], Z.shape[0], 0,
                     gl.GL_ALPHA, gl.GL_UNSIGNED_BYTE, Z )

    # Generate display lists
    dx, dy = width/float(Z.shape[1]), height/float(Z.shape[0])
    base = gl.glGenLists(8*16)
    for i in range(8*16):
        c = chr(i)
        x = i%16
        y = i//16-2
        gl.glNewList(base+i, gl.GL_COMPILE)
        if (c == '\n'):
            gl.glPopMatrix( )
            gl.glTranslatef( 0, -height, 0 )
            gl.glPushMatrix( )
        elif (c == '\t'):
            gl.glTranslatef( 4*width, 0, 0 )
        elif (i >= 32):
            gl.glBegin( gl.GL_QUADS )
            gl.glTexCoord2f( (x  )*dx, (y+1)*dy ), gl.glVertex( 0,     -height )
            gl.glTexCoord2f( (x  )*dx, (y  )*dy ), gl.glVertex( 0,     0 )
            gl.glTexCoord2f( (x+1)*dx, (y  )*dy ), gl.glVertex( width, 0 )
            gl.glTexCoord2f( (x+1)*dx, (y+1)*dy ), gl.glVertex( width, -height )
            gl.glEnd( )
            gl.glTranslatef( width, 0, 0 )
        gl.glEndList( )


if __name__ == '__main__':
    import sys
    glut.glutInit( sys.argv )
    glut.glutInitDisplayMode( glut.GLUT_DOUBLE | glut.GLUT_RGB | glut.GLUT_DEPTH )
    glut.glutCreateWindow( "Freetype OpenGL" )
    glut.glutReshapeWindow( 600, 100 )
    glut.glutDisplayFunc( on_display )
    glut.glutReshapeFunc( on_reshape )
    glut.glutKeyboardFunc( on_keyboard )
    gl.glTexEnvf( gl.GL_TEXTURE_ENV, gl.GL_TEXTURE_ENV_MODE, gl.GL_MODULATE )
    gl.glEnable( gl.GL_DEPTH_TEST )
    gl.glEnable( gl.GL_BLEND )
    gl.glEnable( gl.GL_COLOR_MATERIAL )
    gl.glColorMaterial( gl.GL_FRONT_AND_BACK, gl.GL_AMBIENT_AND_DIFFUSE )
    gl.glBlendFunc( gl.GL_SRC_ALPHA, gl.GL_ONE_MINUS_SRC_ALPHA )
    gl.glEnable( gl.GL_TEXTURE_2D )
    makefont( './VeraMono.ttf', 64 )
    glut.glutMainLoop( )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
#
#  FreeType high-level python API - Copyright 2011-2015 Nicolas P. Rougier
#  Distributed under the terms of the new BSD license.
#
# -----------------------------------------------------------------------------
from __future__ import print_function
from __future__ import division
import sys
from freetype import *

verbose     = 0
debug       = 0
name_tables = 0

def usage( execname ):
    print( )
    print( "ftdump: simple font dumper -- part of the FreeType project" )
    print( "----------------------------------------------------------" )
    print( "Usage: %s [options] fontname", execname )
    print( )
    print( "  -n        print SFNT name tables" )
    print( "  -v        be verbose" )
    print( )
    sys.exit()


def Print_Name( face ):
    print( "font name entries" );
    print( "   family:     %s" % face.family_name.decode("ascii", errors='ignore') )
    print( "   style:      %s" % face.style_name.decode("ascii", errors='ignore') )
    ps_name = face.postscript_name or "UNAVAILABLE"
    print( "   postscript: %s" % ps_name.decode("ascii", errors='ignore') )


def Print_Type( face ):

    print( "font type entries" )

    #   module = &face->driver->root;
    #   printf( "   FreeType driver: %s\n", module->clazz->module_name );

    #  Is it better to dump all sfnt tag names?
    print( "   sfnt wrapped:    ",end="")
    if face.is_sfnt: print( "yes")
    else:            print( "no")

    # is scalable ?
    print( "   type:            ", end="")
    if face.is_scalable:
        print( "scalable, ", end="")
        if face.has_multiple_masters:
            print( "multiple_masters, ", end="")
    if face.has_fixed_sizes:
        print( "fixed size",end="")
    print()

    #  Direction
    print( "   direction:       ", end="" )
    if face.has_horizontal:
        print( "horizontal, ", end="")
    if face.has_vertical:
        print( "vertical", end="")
    print( )

    # Fixed width
    print( "   fixed width:     ", end="")
    if face.is_fixed_width: print( "yes")
    else:                   print( "no")

    # Glyph names
    print( "   glyph names:     ", end="")
    if face.has_glyph_names: print( "yes")
    else:                   print( "no")

    if face.is_scalable:
        print( "   EM size:         %d" % face.units_per_EM )
        print( "   global BBox:     (%ld,%ld):(%ld,%ld)" %
               (face.bbox.xMin, face.bbox.yMin,
                face.bbox.xMax, face.bbox.yMax ))
        print( "   ascent:          %d" % face.ascender )
        print( "   descent:         %d" % face.descender )
        print( "   text height:     %d" % face.height )


def get_platform_id( platform_id ):
    if platform_id == TT_PLATFORM_APPLE_UNICODE:
        return "Apple (Unicode)"
    elif platform_id == TT_PLATFORM_MACINTOSH:
        return "Macintosh"
    elif platform_id == TT_PLATFORM_ISO:
        return "ISO (deprecated)"
    elif platform_id == TT_PLATFORM_MICROSOFT:
        return "Microsoft"
    elif platform_id == TT_PLATFORM_CUSTOM:
        return "custom"
    elif platform_id == TT_PLATFORM_ADOBE:
        return "Adobe"
    else:
      return "UNKNOWN"

def get_name_id( name_id ):
    if name_id == TT_NAME_ID_COPYRIGHT:
        return "copyright"
    elif name_id == TT_NAME_ID_FONT_FAMILY:
        return "font family"
    elif name_id == TT_NAME_ID_FONT_SUBFAMILY:
        return "font subfamily"
    elif name_id == TT_NAME_ID_UNIQUE_ID:
        return "unique ID"
    elif name_id == TT_NAME_ID_FULL_NAME:
        return "full name"
    elif name_id == TT_NAME_ID_VERSION_STRING:
        return "version string"
    elif name_id == TT_NAME_ID_PS_NAME:
        return "PostScript name"
    elif name_id == TT_NAME_ID_TRADEMARK:
        return "trademark"

    # the following values are from the OpenType spec
    elif name_id == TT_NAME_ID_MANUFACTURER:
        return "manufacturer"
    elif name_id == TT_NAME_ID_DESIGNER:
        return "designer"
    elif name_id == TT_NAME_ID_DESCRIPTION:
        return "description"
    elif name_id == TT_NAME_ID_VENDOR_URL:
        return "vendor URL"
    elif name_id == TT_NAME_ID_DESIGNER_URL:
        return "designer URL"
    elif name_id == TT_NAME_ID_LICENSE:
        return "license"
    elif name_id == TT_NAME_ID_LICENSE_URL:
        return "license URL"
    # number 15 is reserved
    elif name_id == TT_NAME_ID_PREFERRED_FAMILY:
        return "preferred family"
    elif name_id == TT_NAME_ID_PREFERRED_SUBFAMILY:
        return "preferred subfamily"
    elif name_id == TT_NAME_ID_MAC_FULL_NAME:
        return "Mac full name"

    # The following code is new as of 2000-01-21
    elif name_id == TT_NAME_ID_SAMPLE_TEXT:
      return "sample text"

    # This is new in OpenType 1.3
    elif name_id == TT_NAME_ID_CID_FINDFONT_NAME:
        return "CID 'findfont' name"
    else:
        return "UNKNOWN";


def Print_Sfnt_Names( face ):
    print( "font string entries" );

    for i in range(face.sfnt_name_count):

        name = face.get_sfnt_name(i)
        print( "   %-15s [%s]" % ( get_name_id( name.name_id ),
                                   get_platform_id( name.platform_id )),end="")

        if name.platform_id == TT_PLATFORM_APPLE_UNICODE:
            if name.encoding_id in [TT_APPLE_ID_DEFAULT,
                                    TT_APPLE_ID_UNICODE_1_1,
                                    TT_APPLE_ID_ISO_10646,
                                    TT_APPLE_ID_UNICODE_2_0]:
                print(name.string.decode('utf-16be', 'ignore'))
            else:
                print( "{unsupported encoding %d}" % name.encoding_id )

        elif name.platform_id == TT_PLATFORM_MACINTOSH:
            if name.language_id != TT_MAC_LANGID_ENGLISH:
                print( " (language=%d)" % name.language_id )
            print ( " : " )
            if name.encoding_id == TT_MAC_ID_ROMAN:
                # FIXME: convert from MacRoman to ASCII/ISO8895-1/whatever
                # (MacRoman is mostly like ISO8895-1 but there are differences)
                print(name.string)
            else:
                print( "{unsupported encoding %d}" % name.encoding_id )

        elif name.platform_id == TT_PLATFORM_ISO:
            if name.encoding_id in [ TT_ISO_ID_7BIT_ASCII,
                                     TT_ISO_ID_8859_1]:
                print(name.string)
            print ( " : " )
            if name.encoding_id == TT_ISO_ID_10646:
                print(name.string.decode('utf-16be', 'ignore'))
            else:
                print( "{unsupported encoding %d}" % name.encoding_id )

        elif name.platform_id == TT_PLATFORM_MICROSOFT:
            if name.language_id != TT_MS_LANGID_ENGLISH_UNITED_STATES:
                print( " (language=0x%04x)" % name.language_id );
            print( " : " )
            if name.encoding_id in [TT_MS_ID_SYMBOL_CS,
                                    TT_MS_ID_UNICODE_CS]:
                print(name.string.decode('utf-16be', 'ignore'))
            else:
                print( "{unsupported encoding %d}" % name.encoding_id )
        else:
           print( "{unsupported platform}" )

        print( )


def Print_Fixed( face ):

    # num_fixed_size
    print( "fixed size\n" )

    # available size
    for i,bsize in enumerate(face.available_sizes):
        print( "   %3d: height %d, width %d" %
               ( i, bsize.height, bsize.width ) )
        print( "        size %.3f, x_ppem %.3f, y_ppem %.3f" %
               ( bsize.size / 64.0,
               bsize.x_ppem / 64.0, bsize.y_ppem / 64.0 ) )


def Print_Charmaps( face ):
    global verbose
    active = -1
    if face.charmap:
        active = face.charmap.index

    # CharMaps
    print( "charmaps" )
    for i,charmap in enumerate(face.charmaps):
        print( "   %d: platform %d, encoding %d, language %d" %
               (i, charmap.platform_id, charmap.encoding_id,
               int(charmap.cmap_language_id)), end="" )
        if i == active:
            print( " (active)", end="" )
        print ( )
        if verbose:
            face.set_charmap( charmap )
            charcode, gindex = face.get_first_char()
            while ( gindex ):
                print( "      0x%04lx => %d" % (charcode, gindex) )
                charcode, gindex = face.get_next_char( charcode, gindex )



# -----------------------------------------------------------------------------
if __name__ == '__main__':
    import getopt
    execname = sys.argv[0]

    if len(sys.argv) < 2:
        usage( execname )

    try:
        opts, args = getopt.getopt(sys.argv[1:], ':nv')
    except getopt.GetoptError:
        usage( execname )

    verbose = False
    name_tables = False

    for o, a in opts:
        if o == "-v":   verbose = True
        elif o == "-n": name_tables = True
        else:           usage( execname )


    face = Face(args[0])
    num_faces = face.num_faces

    if num_faces > 1:
        print( "There are %d faces in this file." % num_faces)
    else:
        print( "There is 1 face in this file.")

    for i in range(num_faces):
        face = Face(args[0], i)

        print( "\n----- Face number: %d -----\n" % i )
        Print_Name( face )
        print( "" )
        Print_Type( face )
        print( "   glyph count:     %d" % face.num_glyphs )

        if name_tables and face.is_sfnt:
            print( )
            Print_Sfnt_Names( face )

        if face.num_fixed_sizes:
            print(  )
            Print_Fixed( face )

        if face.num_charmaps:
            print(  )
            Print_Charmaps( face )




#!/usr/bin/env python
# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
#
#  FreeType high-level python API - Copyright 2011-2015 Nicolas P. Rougier
#  Distributed under the terms of the new BSD license.
#
# -----------------------------------------------------------------------------
from freetype import *

if __name__ == '__main__':
    import numpy
    import matplotlib.pyplot as plt

    face = Face('./Vera.ttf')
    text = 'Hello World !'
    fontsize = 48
    face.set_char_size( fontsize*64 )
    slot = face.glyph

    # First pass to compute bbox
    width = 0
    top = -fontsize #The distance from baseline to the top row of the bbox
    bot = fontsize #The distance from baseline to the bottom row of the bbox (may be negative)
    previous = 0
    for i,c in enumerate(text):
        face.load_char(c)
        bitmap = slot.bitmap
        top = max(top, slot.bitmap_top)
        bot = min(bot, slot.bitmap_top - bitmap.rows)
        kerning = face.get_kerning(previous, c)
        width += (slot.advance.x >> 6) + (kerning.x >> 6)
        previous = c

    height = top - bot
    Z = numpy.zeros((height,width), dtype=numpy.ubyte)

    # Second pass for actual rendering
    x, y = 0, 0
    previous = 0
    for c in text:
        face.load_char(c)
        bitmap = slot.bitmap
        w,h = bitmap.width, bitmap.rows
        y = top - slot.bitmap_top
        kerning = face.get_kerning(previous, c)
        x += (kerning.x >> 6)
        Z[y:y+h,x+left:x+left+w] += numpy.array(bitmap.buffer, dtype='ubyte').reshape(h,w)
        x += (slot.advance.x >> 6)
        previous = c

    plt.figure(figsize=(10, 10*Z.shape[0]/float(Z.shape[1])))
    plt.imshow(Z, interpolation='nearest', origin='upper', cmap=plt.cm.gray)
    plt.xticks([]), plt.yticks([])
    plt.show()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
#
#  Variable Font example - Copyright 2020 Josh Hadley, based on hello-world.py
#  Distributed under the terms of the new BSD license.
#
# -----------------------------------------------------------------------------
from freetype import *

if __name__ == '__main__':
    import numpy
    import matplotlib.pyplot as plt

    face = Face('./SourceSansVariable-Roman.otf')
    text = 'Hello variable fonts!'
    face.set_char_size( 48*64 )
    slot = face.glyph
    face.set_var_design_coords((900,))  # set to widest for width calc.

    # First pass to compute bbox
    width, height, baseline = 0, 0, 0
    previous = 0
    for i,c in enumerate(text):
        face.load_char(c)
        bitmap = slot.bitmap
        height = max(height,
                     bitmap.rows + max(0,-(slot.bitmap_top-bitmap.rows)))
        baseline = max(baseline, max(0,-(slot.bitmap_top-bitmap.rows)))
        kerning = face.get_kerning(previous, c)
        width += (slot.advance.x >> 6) + (kerning.x >> 6)
        previous = c

    height *= 12

    Z = numpy.zeros((height,width), dtype=numpy.ubyte)

    # Second pass for actual rendering. Iterate through some weight axis values
    # and render the result, one per line.
    for i, wght in enumerate((900, 750, 675, 600, 500, 450, 325, 275, 100)):
        face.set_var_design_coords((wght,)) # set the weight axis value using "design coords"
        x, y = 0, 0
        previous = 0
        for c in text:
            face.load_char(c)
            bitmap = slot.bitmap
            top = slot.bitmap_top
            left = slot.bitmap_left
            w,h = bitmap.width, bitmap.rows
            y = (height - baseline - top) - (i * 48)
            kerning = face.get_kerning(previous, c)
            x += (kerning.x >> 6)
            Z[y:y+h,x+left:x+left+w] += numpy.array(bitmap.buffer, dtype='ubyte').reshape(h,w)
            x += (slot.advance.x >> 6)
            previous = c

    plt.figure(figsize=(10, 10*Z.shape[0]/float(Z.shape[1])))
    plt.imshow(Z, interpolation='nearest', origin='upper', cmap=plt.cm.gray)
    plt.xticks([]), plt.yticks([])
    plt.show()



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
#
#  FreeType high-level python API - Copyright 2011-2015 Nicolas P. Rougier
#  Distributed under the terms of the new BSD license.
#
# -----------------------------------------------------------------------------
"""
A list of bit-field constants used within the 'flags' field of the
FT_Open_Args structure.


FT_OPEN_MEMORY

  This is a memory-based stream.


FT_OPEN_STREAM

  Copy the stream from the 'stream' field.


FT_OPEN_PATHNAME

  Create a new input stream from a C path name.


FT_OPEN_DRIVER

  Use the 'driver' field.


FT_OPEN_PARAMS

  Use the 'num_params' and 'params' fields.
"""
FT_OPEN_MODES = {'FT_OPEN_MEMORY':   0x1,
                 'FT_OPEN_STREAM':   0x2,
                 'FT_OPEN_PATHNAME': 0x4,
                 'FT_OPEN_DRIVER':   0x8,
                 'FT_OPEN_PARAMS':   0x10 }
globals().update(FT_OPEN_MODES)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  Identify the color glyph format.
#
#  Copyright 2023 Hin-Tak Leung
#  Distributed under the terms of the new BSD license.
#
#  Note:
#      This is a per-glyph property.
#
#      Glyph 1426 in SEGUIEMJ.TTF is a glyph layer of glyph 1425.
#      Glyph 1426 is a plain outline, 1425 is a COLRv0 glyph.
#
#      Noto Color Emoji is only valid for size 109.
#
#      Apple Color Emoji works for size 160 and a few other specific sizes.
#
#      SVG and COLRv1 is not exclusive - some fonts have both.

from freetype import *

if __name__ == '__main__':
    import sys
    execname = sys.argv[0]

    if len(sys.argv) < 3:
        print("Example usage: %s TrajanColor-Concept.otf <glyph_id> [size]" % execname)
        exit(1)

    face = Face(sys.argv[1])

    glyph_id = c_uint(int(sys.argv[2]))

    if (len(sys.argv) > 3):
        size = int(sys.argv[3])
    else:
        size = 160

    face.set_char_size( size*64 )
    face.load_glyph( glyph_id, FT_LOAD_COLOR )

    print("glyph id %d is " % (glyph_id.value), end='')
    
    if (face.glyph.format == FT_GLYPH_FORMAT_SVG):
        print("SVG; ", end='')
    if (face.glyph.format == FT_GLYPH_FORMAT_BITMAP):
        print("Bitmap; ", end='')
    if (face.glyph.format == FT_GLYPH_FORMAT_OUTLINE):
        print("Outline; ", end='')

    opaqueLayerPaint = FT_OpaquePaint(None, 1)
    if (FT_Get_Color_Glyph_Paint(face._FT_Face, glyph_id,
                                 FT_COLOR_INCLUDE_ROOT_TRANSFORM,
                                 byref(opaqueLayerPaint))):
        print("COLRv1; ", end='')
    else:
        layerGlyphIndex = FT_UInt(0)
        layerColorIndex = FT_UInt(0)
        layerIterator   = FT_LayerIterator( 0, 0, None )
        layer_count = 0
        if (FT_Get_Color_Glyph_Layer(face._FT_Face, glyph_id,
                                     byref(layerGlyphIndex),
                                     byref(layerColorIndex),
                                     byref(layerIterator))):
            if (layerGlyphIndex.value != 0):
                print("COLRv0; ", end='')

    print ("") # line ending



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
#
#  FreeType high-level python API - Copyright 2011-2015 Nicolas P. Rougier
#  Distributed under the terms of the new BSD license.
#
# -----------------------------------------------------------------------------
"""
A list of bit-flags used to indicate the style of a given face. These are
used in the 'style_flags' field of FT_FaceRec.


FT_STYLE_FLAG_ITALIC

  Indicates that a given face style is italic or oblique.


FT_STYLE_FLAG_BOLD

  Indicates that a given face is bold.
"""
FT_STYLE_FLAGS = {'FT_STYLE_FLAG_ITALIC' : 1,
                   'FT_STYLE_FLAG_BOLD'   : 2 }
globals().update(FT_STYLE_FLAGS)



# -----------------------------------------------------------------------------
#  FreeType high-level python API - copyright 2011 Nicolas P. Rougier
#  Distributed under the terms of the new BSD license.
# -----------------------------------------------------------------------------

import distutils
import distutils.dir_util
import distutils.file_util
import os
import subprocess
import sys
from io import open
from os import path

from setuptools import setup

if os.environ.get("FREETYPEPY_BUNDLE_FT"):
    print("# Will build and bundle FreeType.")

    from setuptools import Extension
    from setuptools.command.build_ext import build_ext
    from wheel.bdist_wheel import bdist_wheel

    class BdistWheel(bdist_wheel):
        def get_tag(self):
            return ('py3', 'none') + bdist_wheel.get_tag(self)[2:]

    class SharedLibrary(Extension):
        """Object that describes the library (filename) and how to make it."""
        if sys.platform == "darwin":
            suffix = ".dylib"
        elif sys.platform == "win32":
            suffix = ".dll"
        else:
            suffix = ".so"

        def __init__(self, name, cmd, cwd=".", output_dir=".", env=None):
            Extension.__init__(self, name, sources=[])
            self.cmd = cmd
            self.cwd = path.normpath(cwd)
            self.output_dir = path.normpath(output_dir)
            self.env = env or dict(os.environ)

    class SharedLibBuildExt(build_ext):
        """Object representing command to produce and install a shared
        library."""

        # Needed to make setuptools and wheel believe they're looking at an
        # extension instead of a shared library.
        def get_ext_filename(self, ext_name):
            for ext in self.extensions:
                if isinstance(ext, SharedLibrary):
                    return os.path.join(*ext_name.split('.')) + ext.suffix
            return build_ext.get_ext_filename(self, ext_name)

        def build_extension(self, ext):
            if not isinstance(ext, SharedLibrary):
                build_ext.build_extension(self, ext)
                return

            distutils.log.info("running '{}'".format(ext.cmd))
            if not self.dry_run:
                rv = subprocess.Popen(
                    ext.cmd, cwd=ext.cwd, env=ext.env, shell=True).wait()
                if rv != 0:
                    sys.exit(rv)

            lib_name = ext.name.split(".")[-1] + ext.suffix
            lib_fullpath = path.join(ext.output_dir, lib_name)
            dest_path = self.get_ext_fullpath(ext.name)

            distutils.dir_util.mkpath(
                path.dirname(dest_path),
                verbose=self.verbose,
                dry_run=self.dry_run)

            distutils.file_util.copy_file(
                lib_fullpath,
                dest_path,
                verbose=self.verbose,
                dry_run=self.dry_run)

    ext_modules = [
        SharedLibrary(
            "freetype.libfreetype",  # package.shared_lib_name
            cmd='"{}" ./setup-build-freetype.py'.format(sys.executable),
            output_dir="build/local/lib")
    ]
    cmdclass = {
        'bdist_wheel': BdistWheel,
        'build_ext': SharedLibBuildExt
    }

else:
    print("# Will use the system-provided FreeType.")
    ext_modules = []
    cmdclass = {}

description = open(
    path.join(path.abspath(path.dirname(__file__)), 'README.rst'),
    encoding='utf-8').read()

setup(
    name='freetype-py',
    use_scm_version=True,
    description='Freetype python bindings',
    long_description=description,
    author='Nicolas P. Rougier',
    author_email='Nicolas.Rougier@inria.fr',
    url='https://github.com/rougier/freetype-py',
    python_requires=">=3.7",
    packages=['freetype', 'freetype.ft_enums', 'freetype.__pyinstaller', 'freetype.__pyinstaller.tests'],
    ext_modules=ext_modules,
    zip_safe=False if ext_modules else True,
    cmdclass=cmdclass,
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: X11 Applications',
        'Environment :: MacOS X',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: MacOS',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: POSIX',
        'Operating System :: Unix',
        'Programming Language :: Python :: 3',
        'Topic :: Multimedia :: Graphics',
    ],
    keywords=['freetype', 'font'],
    setup_requires=['setuptools_scm'],
    entry_points={
        "pyinstaller40": [
            "hook-dirs = freetype.__pyinstaller:get_hook_dirs",
            "tests = freetype.__pyinstaller:get_test_dirs",
        ],
    },
)



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
#  FreeType high-level python API - Copyright 2011-2015 Nicolas P. Rougier
#  Distributed under the terms of the new BSD license.
# -----------------------------------------------------------------------------
'''
Freetype raw API

This is the raw ctypes freetype binding.
'''
import os
import platform
from ctypes import *
import ctypes.util

import freetype
from freetype.ft_types import *
from freetype.ft_enums import *
from freetype.ft_errors import *
from freetype.ft_structs import *

# First, look for a bundled FreeType shared object on the top-level of the
# installed freetype-py module.
system = platform.system()
if system == 'Windows':
    library_name = 'libfreetype.dll'
elif system == 'Darwin':
    library_name = 'libfreetype.dylib'
else:
    library_name = 'libfreetype.so'

filename = os.path.join(os.path.dirname(freetype.__file__), library_name)

# If no bundled shared object is found, look for a system-wide installed one.
if not os.path.exists(filename):
    # on windows all ctypes does when checking for the library
    # is to append .dll to the end and look for an exact match
    # within any entry in PATH.
    filename = ctypes.util.find_library('freetype')

    if filename is None:
        if platform.system() == 'Windows':
            # Check current working directory for dll as ctypes fails to do so
            filename = os.path.join(os.path.realpath('.'), "freetype.dll")
        else:
            filename = library_name

try:
    _lib = ctypes.CDLL(filename)
except (OSError, TypeError):
    _lib = None
    raise RuntimeError('Freetype library not found')

FT_Init_FreeType       = _lib.FT_Init_FreeType
FT_Done_FreeType       = _lib.FT_Done_FreeType
FT_Library_Version     = _lib.FT_Library_Version

try:
    FT_Library_SetLcdFilter= _lib.FT_Library_SetLcdFilter
except AttributeError:
    def FT_Library_SetLcdFilter (*args, **kwargs):
        return 0
try:
    FT_Library_SetLcdFilterWeights = _lib.FT_Library_SetLcdFilterWeights
except AttributeError:
    pass

FT_New_Face            = _lib.FT_New_Face
FT_New_Memory_Face     = _lib.FT_New_Memory_Face
FT_Open_Face           = _lib.FT_Open_Face
FT_Attach_File         = _lib.FT_Attach_File
FT_Attach_Stream       = _lib.FT_Attach_Stream

try:
    FT_Reference_Face      = _lib.FT_Reference_Face
except AttributeError:
    pass

FT_Done_Face           = _lib.FT_Done_Face
FT_Done_Glyph          = _lib.FT_Done_Glyph
FT_Select_Size         = _lib.FT_Select_Size
FT_Request_Size        = _lib.FT_Request_Size
FT_Set_Char_Size       = _lib.FT_Set_Char_Size
FT_Set_Pixel_Sizes     = _lib.FT_Set_Pixel_Sizes
FT_Load_Glyph          = _lib.FT_Load_Glyph
FT_Load_Char           = _lib.FT_Load_Char
FT_Set_Transform       = _lib.FT_Set_Transform
FT_Render_Glyph        = _lib.FT_Render_Glyph
FT_Get_Kerning         = _lib.FT_Get_Kerning
FT_Get_Track_Kerning   = _lib.FT_Get_Track_Kerning
FT_Get_Glyph_Name      = _lib.FT_Get_Glyph_Name
FT_Get_Glyph           = _lib.FT_Get_Glyph

FT_Glyph_Get_CBox      = _lib.FT_Glyph_Get_CBox

FT_Get_Postscript_Name = _lib.FT_Get_Postscript_Name
FT_Get_Postscript_Name.restype = c_char_p
FT_Select_Charmap      = _lib.FT_Select_Charmap
FT_Set_Charmap         = _lib.FT_Set_Charmap
FT_Get_Charmap_Index   = _lib.FT_Get_Charmap_Index
FT_Get_CMap_Language_ID= _lib.FT_Get_CMap_Language_ID
try:
    #introduced between 2.2.x and 2.3.x
    FT_Get_CMap_Format     = _lib.FT_Get_CMap_Format
except AttributeError:
    pass
FT_Get_Char_Index      = _lib.FT_Get_Char_Index
FT_Get_First_Char      = _lib.FT_Get_First_Char
FT_Get_Next_Char       = _lib.FT_Get_Next_Char
FT_Get_Name_Index      = _lib.FT_Get_Name_Index
FT_Get_SubGlyph_Info   = _lib.FT_Get_SubGlyph_Info

try:
    FT_Get_FSType_Flags    = _lib.FT_Get_FSType_Flags
    FT_Get_FSType_Flags.restype  = c_ushort
except AttributeError:
    pass

FT_Get_X11_Font_Format = _lib.FT_Get_X11_Font_Format
FT_Get_X11_Font_Format.restype = c_char_p

FT_Get_Sfnt_Name_Count = _lib.FT_Get_Sfnt_Name_Count
FT_Get_Sfnt_Name       = _lib.FT_Get_Sfnt_Name
try:
    # introduced between 2.2.x and 2.3.x
    FT_Get_Advance         = _lib.FT_Get_Advance
except AttributeError:
    pass


FT_Outline_GetInsideBorder  = _lib.FT_Outline_GetInsideBorder
FT_Outline_GetOutsideBorder = _lib.FT_Outline_GetOutsideBorder
FT_Outline_Get_BBox         = _lib.FT_Outline_Get_BBox
FT_Outline_Get_CBox         = _lib.FT_Outline_Get_CBox
try:
    # since 2.4.10
    FT_Outline_EmboldenXY       = _lib.FT_Outline_EmboldenXY
except AttributeError:
    pass
FT_Stroker_New              = _lib.FT_Stroker_New
FT_Stroker_Set              = _lib.FT_Stroker_Set
FT_Stroker_Rewind           = _lib.FT_Stroker_Rewind
FT_Stroker_ParseOutline     = _lib.FT_Stroker_ParseOutline
FT_Stroker_BeginSubPath     = _lib.FT_Stroker_BeginSubPath
FT_Stroker_EndSubPath       = _lib.FT_Stroker_EndSubPath
FT_Stroker_LineTo           = _lib.FT_Stroker_LineTo
FT_Stroker_ConicTo          = _lib.FT_Stroker_ConicTo
FT_Stroker_CubicTo          = _lib.FT_Stroker_CubicTo
FT_Stroker_GetBorderCounts  = _lib.FT_Stroker_GetBorderCounts
FT_Stroker_ExportBorder     = _lib.FT_Stroker_ExportBorder
FT_Stroker_GetCounts        = _lib.FT_Stroker_GetCounts
FT_Stroker_Export           = _lib.FT_Stroker_Export
FT_Stroker_Done             = _lib.FT_Stroker_Done
FT_Glyph_Stroke             = _lib.FT_Glyph_Stroke
FT_Glyph_StrokeBorder       = _lib.FT_Glyph_StrokeBorder
FT_Glyph_To_Bitmap          = _lib.FT_Glyph_To_Bitmap


# FT_Property_Get/FT_Property_Set requires FreeType 2.7.x+
try:
    FT_Property_Get    = _lib.FT_Property_Get
    FT_Property_Set    = _lib.FT_Property_Set
except AttributeError:
    pass

# These two are only found when TT debugger is enabled
try:
    TT_New_Context     = _lib.TT_New_Context
    TT_RunIns          = _lib.TT_RunIns
except AttributeError:
    pass


# Routines for variable font support. These were introduced at different
# points in FreeType's history (except for FT_Get_MM_Var which has been
# around for a long time).
FT_Get_MM_Var  = _lib.FT_Get_MM_Var  # v2.old

# -- since 2.8.1 for sure (some 2.7.1 or possibly older, but to be safe,
# implementation should check FT version >= 2.8.1
try:
    FT_Get_Var_Axis_Flags         = _lib.FT_Get_Var_Axis_Flags
    FT_Get_Var_Blend_Coordinates  = _lib.FT_Get_Var_Blend_Coordinates
    FT_Get_Var_Design_Coordinates = _lib.FT_Get_Var_Design_Coordinates
    FT_Set_Var_Blend_Coordinates  = _lib.FT_Set_Var_Blend_Coordinates
    FT_Set_Var_Design_Coordinates = _lib.FT_Set_Var_Design_Coordinates
except AttributeError:
    pass

# -- since v2.9; we can work around if these are not present.
try:
    FT_Done_MM_Var        = _lib.FT_Done_MM_Var
    FT_Set_Named_Instance = _lib.FT_Set_Named_Instance
except AttributeError:
    pass

# Wholesale import of 102 routines which can be reasonably expected
# to be found in freetype 2.2.x onwards. Some of these might need
# to be protected with try:/except AttributeError: in some freetype builds.

FTC_CMapCache_Lookup           = _lib.FTC_CMapCache_Lookup
FTC_CMapCache_New              = _lib.FTC_CMapCache_New
FTC_ImageCache_Lookup          = _lib.FTC_ImageCache_Lookup
FTC_ImageCache_New             = _lib.FTC_ImageCache_New
FTC_Manager_Done               = _lib.FTC_Manager_Done
FTC_Manager_LookupFace         = _lib.FTC_Manager_LookupFace
FTC_Manager_LookupSize         = _lib.FTC_Manager_LookupSize
FTC_Manager_New                = _lib.FTC_Manager_New
FTC_Manager_RemoveFaceID       = _lib.FTC_Manager_RemoveFaceID
FTC_Manager_Reset              = _lib.FTC_Manager_Reset
FTC_Node_Unref                 = _lib.FTC_Node_Unref
FTC_SBitCache_Lookup           = _lib.FTC_SBitCache_Lookup
FTC_SBitCache_New              = _lib.FTC_SBitCache_New

FT_Activate_Size               = _lib.FT_Activate_Size
FT_Add_Default_Modules         = _lib.FT_Add_Default_Modules
FT_Add_Module                  = _lib.FT_Add_Module
FT_Angle_Diff                  = _lib.FT_Angle_Diff
FT_Atan2                       = _lib.FT_Atan2
FT_Bitmap_Convert              = _lib.FT_Bitmap_Convert
FT_Bitmap_Copy                 = _lib.FT_Bitmap_Copy
FT_Bitmap_Done                 = _lib.FT_Bitmap_Done
FT_Bitmap_Embolden             = _lib.FT_Bitmap_Embolden
FT_Bitmap_New                  = _lib.FT_Bitmap_New
FT_CeilFix                     = _lib.FT_CeilFix
FT_ClassicKern_Free            = _lib.FT_ClassicKern_Free
FT_ClassicKern_Validate        = _lib.FT_ClassicKern_Validate
FT_Cos                         = _lib.FT_Cos
FT_DivFix                      = _lib.FT_DivFix
FT_Done_Library                = _lib.FT_Done_Library
FT_Done_Size                   = _lib.FT_Done_Size
FT_FloorFix                    = _lib.FT_FloorFix
try:
    # Not in default windows build of 2.8.x
    FT_Get_BDF_Charset_ID          = _lib.FT_Get_BDF_Charset_ID
    FT_Get_BDF_Property            = _lib.FT_Get_BDF_Property
except AttributeError:
    pass
try:
    FT_Get_Color_Glyph_Layer           = _lib.FT_Get_Color_Glyph_Layer # 2.10
    FT_Get_Color_Glyph_Layer.restype   = FT_Bool
    FT_Get_Color_Glyph_Layer.argtypes  = [FT_Face, FT_UInt, POINTER(FT_UInt), POINTER(FT_UInt), POINTER(FT_LayerIterator)]
    FT_Get_Color_Glyph_Paint           = _lib.FT_Get_Color_Glyph_Paint # 2.13
    FT_Get_Color_Glyph_Paint.restype   = FT_Bool
    FT_Get_Color_Glyph_Paint.argtypes  = [FT_Face, FT_UInt,
                                          FT_UInt, # FT_Color_Root_Transform enums
                                          POINTER(FT_OpaquePaint)]
except AttributeError:
    pass
try:
    FT_Face_GetCharVariantIndex          = _lib.FT_Face_GetCharVariantIndex
    FT_Face_GetCharVariantIndex.argtypes = [FT_Face, FT_ULong, FT_ULong]
    FT_Face_GetCharVariantIndex.restype  = FT_UInt

    FT_Face_GetCharVariantIsDefault          = _lib.FT_Face_GetCharVariantIsDefault
    FT_Face_GetCharVariantIsDefault.argtypes = [FT_Face, FT_ULong, FT_ULong]
    FT_Face_GetCharVariantIsDefault.restype  = FT_Int

    FT_Face_GetVariantSelectors          = _lib.FT_Face_GetVariantSelectors
    FT_Face_GetVariantSelectors.argtypes = [FT_Face]
    FT_Face_GetVariantSelectors.restype  = POINTER(FT_UInt32)

    FT_Face_GetVariantsOfChar          = _lib.FT_Face_GetVariantsOfChar
    FT_Face_GetVariantsOfChar.argtypes = [FT_Face, FT_ULong]
    FT_Face_GetVariantsOfChar.restype  = POINTER(FT_UInt32)

    FT_Face_GetCharsOfVariant          = _lib.FT_Face_GetCharsOfVariant
    FT_Face_GetVariantsOfChar.argtypes = [FT_Face, FT_ULong]
    FT_Face_GetCharsOfVariant.restype  = POINTER(FT_UInt32)
except AttributeError:
    pass
FT_Get_Module                  = _lib.FT_Get_Module
FT_Get_Multi_Master            = _lib.FT_Get_Multi_Master
FT_Get_PFR_Advance             = _lib.FT_Get_PFR_Advance
FT_Get_PFR_Kerning             = _lib.FT_Get_PFR_Kerning
FT_Get_PFR_Metrics             = _lib.FT_Get_PFR_Metrics
FT_Get_PS_Font_Info            = _lib.FT_Get_PS_Font_Info
FT_Get_PS_Font_Private         = _lib.FT_Get_PS_Font_Private
FT_Get_Renderer                = _lib.FT_Get_Renderer
FT_Get_Sfnt_Table              = _lib.FT_Get_Sfnt_Table
FT_Get_TrueType_Engine_Type    = _lib.FT_Get_TrueType_Engine_Type
FT_Get_WinFNT_Header           = _lib.FT_Get_WinFNT_Header
FT_Glyph_Copy                  = _lib.FT_Glyph_Copy
FT_GlyphSlot_Embolden          = _lib.FT_GlyphSlot_Embolden
FT_GlyphSlot_Oblique           = _lib.FT_GlyphSlot_Oblique
FT_GlyphSlot_Own_Bitmap        = _lib.FT_GlyphSlot_Own_Bitmap
FT_Glyph_Transform             = _lib.FT_Glyph_Transform
FT_Has_PS_Glyph_Names          = _lib.FT_Has_PS_Glyph_Names
FT_List_Add                    = _lib.FT_List_Add
FT_List_Finalize               = _lib.FT_List_Finalize
FT_List_Find                   = _lib.FT_List_Find
FT_List_Insert                 = _lib.FT_List_Insert
FT_List_Iterate                = _lib.FT_List_Iterate
FT_List_Remove                 = _lib.FT_List_Remove
FT_List_Up                     = _lib.FT_List_Up
FT_Load_Sfnt_Table             = _lib.FT_Load_Sfnt_Table
FT_Matrix_Invert               = _lib.FT_Matrix_Invert
FT_Matrix_Multiply             = _lib.FT_Matrix_Multiply
FT_MulDiv                      = _lib.FT_MulDiv
FT_MulFix                      = _lib.FT_MulFix
FT_New_Library                 = _lib.FT_New_Library
FT_New_Size                    = _lib.FT_New_Size
FT_OpenType_Free               = _lib.FT_OpenType_Free
FT_OpenType_Validate           = _lib.FT_OpenType_Validate
FT_Outline_Check               = _lib.FT_Outline_Check
FT_Outline_Copy                = _lib.FT_Outline_Copy
FT_Outline_Decompose           = _lib.FT_Outline_Decompose
FT_Outline_Done                = _lib.FT_Outline_Done
FT_Outline_Embolden            = _lib.FT_Outline_Embolden
FT_Outline_Get_Bitmap          = _lib.FT_Outline_Get_Bitmap
FT_Outline_Get_Orientation     = _lib.FT_Outline_Get_Orientation
FT_Outline_New                 = _lib.FT_Outline_New
FT_Outline_Render              = _lib.FT_Outline_Render
FT_Outline_Reverse             = _lib.FT_Outline_Reverse
FT_Outline_Transform           = _lib.FT_Outline_Transform
FT_Outline_Translate           = _lib.FT_Outline_Translate
FT_Remove_Module               = _lib.FT_Remove_Module
FT_Render_Glyph                = _lib.FT_Render_Glyph
FT_RoundFix                    = _lib.FT_RoundFix
FT_Set_Debug_Hook              = _lib.FT_Set_Debug_Hook
FT_Set_MM_Blend_Coordinates    = _lib.FT_Set_MM_Blend_Coordinates
FT_Set_MM_Design_Coordinates   = _lib.FT_Set_MM_Design_Coordinates
FT_Set_Renderer                = _lib.FT_Set_Renderer
FT_Sfnt_Table_Info             = _lib.FT_Sfnt_Table_Info
FT_Sin                         = _lib.FT_Sin
FT_Stream_OpenGzip             = _lib.FT_Stream_OpenGzip
FT_Stream_OpenLZW              = _lib.FT_Stream_OpenLZW
FT_Tan                         = _lib.FT_Tan
FT_TrueTypeGX_Free             = _lib.FT_TrueTypeGX_Free
FT_TrueTypeGX_Validate         = _lib.FT_TrueTypeGX_Validate
FT_Vector_From_Polar           = _lib.FT_Vector_From_Polar
FT_Vector_Length               = _lib.FT_Vector_Length
FT_Vector_Polarize             = _lib.FT_Vector_Polarize
FT_Vector_Rotate               = _lib.FT_Vector_Rotate
FT_Vector_Transform            = _lib.FT_Vector_Transform
FT_Vector_Unit                 = _lib.FT_Vector_Unit

# Wholesale import ends



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
#
#  FreeType high-level python API - Copyright 2011-2015 Nicolas P. Rougier
#  Distributed under the terms of the new BSD license.
#
# -----------------------------------------------------------------------------
"""
A list of valid values for the 'platform_id' identifier code in FT_CharMapRec
and FT_SfntName structures.


TT_PLATFORM_APPLE_UNICODE

  Used by Apple to indicate a Unicode character map and/or name entry. See
  TT_APPLE_ID_XXX for corresponding 'encoding_id' values. Note that name
  entries in this format are coded as big-endian UCS-2 character codes only.


TT_PLATFORM_MACINTOSH

  Used by Apple to indicate a MacOS-specific charmap and/or name entry. See
  TT_MAC_ID_XXX for corresponding 'encoding_id' values. Note that most TrueType
  fonts contain an Apple roman charmap to be usable on MacOS systems (even if
  they contain a Microsoft charmap as well).


TT_PLATFORM_ISO

  This value was used to specify ISO/IEC 10646 charmaps. It is however now
  deprecated. See TT_ISO_ID_XXX for a list of corresponding 'encoding_id'
  values.


TT_PLATFORM_MICROSOFT

  Used by Microsoft to indicate Windows-specific charmaps. See TT_MS_ID_XXX for
  a list of corresponding 'encoding_id' values. Note that most fonts contain a
  Unicode charmap using (TT_PLATFORM_MICROSOFT, TT_MS_ID_UNICODE_CS).


TT_PLATFORM_CUSTOM

  Used to indicate application-specific charmaps.


TT_PLATFORM_ADOBE

  This value isn't part of any font format specification, but is used by
  FreeType to report Adobe-specific charmaps in an FT_CharMapRec structure. See
  TT_ADOBE_ID_XXX.
"""

TT_PLATFORMS = {
    'TT_PLATFORM_APPLE_UNICODE' : 0,
    'TT_PLATFORM_MACINTOSH'     : 1,
    'TT_PLATFORM_ISO'           : 2, # deprecated
    'TT_PLATFORM_MICROSOFT'     : 3,
    'TT_PLATFORM_CUSTOM'        : 4,
    'TT_PLATFORM_ADOBE'         : 7} # artificial
globals().update(TT_PLATFORMS)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
#
#  pycairo/cairocffi-based wordle example - Copyright 2017 Hin-Tak Leung
#  Distributed under the terms of the new BSD license.
#
#  rewrite of the numply,matplotlib-based example from Nicolas P. Rougier
#  - Cairo can paint partly off-screen, so this example does!
#
#  This example behaves differently under pycairo 1.11+ (released in 2017-04-09).
#  Auto-checked. On older pycairo, it does not draw partial patterns at the edges.
#  Also, Suface.get_data() is not in the "python 3, pycairo < 1.11" combination.
#
# -----------------------------------------------------------------------------
from math import cos, sin
from numpy import random, sort, sqrt, ndarray, ubyte
from freetype import *

try:
    # pycairo 1.11+:
    from cairo import Region, RectangleInt, REGION_OVERLAP_OUT
except ImportError:
    # stubs for pycairo < 1.11:
    class Region:
        def __init__(self):
            return
        def union(self, rec):
            return
        def contains_rectangle(self, rec):
            # inhibit drawing
            return True
    class RectangleInt:
        def __init__(self, x, y, w, h):
            return
    REGION_OVERLAP_OUT = False

from cairo import Context, ImageSurface, FORMAT_A8, FORMAT_ARGB32, Matrix
from bitmap_to_surface import make_image_surface

def make_label(text, filename, size=12, angle=0):
    '''
    Parameters:
    -----------
    text : string
        Text to be displayed
    filename : string
        Path to a font
    size : int
        Font size in 1/64th points
    angle : float
        Text angle in degrees
    '''
    face = Face(filename)
    face.set_char_size( size*64 )
    # FT_Angle is a 16.16 fixed-point value expressed in degrees.
    angle = FT_Angle(angle * 65536)
    matrix  = FT_Matrix( FT_Cos( angle ),
                         - FT_Sin( angle ),
                         FT_Sin( angle ) ,
                         FT_Cos( angle ) )
    flags = FT_LOAD_RENDER
    pen = FT_Vector(0,0)
    FT_Set_Transform( face._FT_Face, byref(matrix), byref(pen) )
    previous = 0
    xmin, xmax = 0, 0
    ymin, ymax = 0, 0
    for c in text:
        face.load_char(c, flags)
        kerning = face.get_kerning(previous, c)
        previous = c
        bitmap = face.glyph.bitmap
        pitch  = face.glyph.bitmap.pitch
        width  = face.glyph.bitmap.width
        rows   = face.glyph.bitmap.rows
        top    = face.glyph.bitmap_top
        left   = face.glyph.bitmap_left
        pen.x += kerning.x
        x0 = (pen.x >> 6) + left
        x1 = x0 + width
        y0 = (pen.y >> 6) - (rows - top)
        y1 = y0 + rows
        xmin, xmax = min(xmin, x0),  max(xmax, x1)
        ymin, ymax = min(ymin, y0), max(ymax, y1)
        pen.x += face.glyph.advance.x
        pen.y += face.glyph.advance.y

    L = ImageSurface(FORMAT_A8, xmax-xmin, ymax-ymin)
    previous = 0
    pen.x, pen.y = 0, 0
    ctx = Context(L)
    for c in text:
        face.load_char(c, flags)
        kerning = face.get_kerning(previous, c)
        previous = c
        bitmap = face.glyph.bitmap
        pitch  = face.glyph.bitmap.pitch
        width  = face.glyph.bitmap.width
        rows   = face.glyph.bitmap.rows
        top    = face.glyph.bitmap_top
        left   = face.glyph.bitmap_left
        pen.x += kerning.x
        x = (pen.x >> 6) - xmin + left
        y = - (pen.y >> 6) + ymax - top
        if (width > 0):
            glyph_surface = make_image_surface(face.glyph.bitmap)
            ctx.set_source_surface(glyph_surface, x, y)
            ctx.paint()
        pen.x += face.glyph.advance.x
        pen.y += face.glyph.advance.y

    L.flush()
    return L


if __name__ == '__main__':
    from PIL import Image

    n_words = 200
    H, W, dpi = 600, 800, 72.0
    I = ImageSurface(FORMAT_A8, W, H)
    ctxI = Context(I)
    ctxI.rectangle(0,0,800,600)
    ctxI.set_source_rgba (0.9, 0.9, 0.9, 0)
    ctxI.fill()
    S = random.normal(0,1,n_words)
    S = (S-S.min())/(S.max()-S.min())
    S = sort(1-sqrt(S))[::-1]
    sizes = (12 + S*48).astype(int).tolist()

    def spiral():
        eccentricity = 1.5
        radius = 8
        step = 0.1
        t = 0
        while True:
            t += step
            yield eccentricity*radius*t*cos(t), radius*t*sin(t)

    drawn_regions = Region()
    for size in sizes:
        angle = random.randint(-25,25)
        try:
            L = make_label('Hello', './Vera.ttf', size, angle=angle)
        except NotImplementedError:
            raise SystemExit("For python 3.x, you need pycairo >= 1.11+ (from https://github.com/pygobject/pycairo)")
        h = L.get_height()
        w = L.get_width()
        if h < H and w < W:
            x0 = W//2 + (random.uniform()-.1)*50
            y0 = H//2 + (random.uniform()-.1)*50
            for dx,dy in spiral():
                c = .25+.75*random.random()
                x = int(x0+dx)
                y = int(y0+dy)
                checked = False
                I.flush()
                if not (x <= w//2 or y <= h//2 or x >= (W-w//2) or y >= (H-h//2)):
                    ndI = ndarray(shape=(h,w), buffer=I.get_data(), dtype=ubyte, order='C',
                                  offset=(x-w//2) + I.get_stride() * (y-h//2),
                                  strides=[I.get_stride(), 1])
                    ndL = ndarray(shape=(h,w), buffer=L.get_data(), dtype=ubyte, order='C',
                                  strides=[L.get_stride(), 1])
                    if ((ndI * ndL).sum() == 0):
                       checked = True
                new_region = RectangleInt(x-w//2, y-h//2, w, h)
                if  (checked or ( drawn_regions.contains_rectangle(new_region) == REGION_OVERLAP_OUT )):
                    ctxI.set_source_surface(L, 0, 0)
                    pattern = ctxI.get_source()
                    scalematrix = Matrix()
                    scalematrix.scale(1.0,1.0)
                    scalematrix.translate(w//2 - x, h//2 - y)
                    pattern.set_matrix(scalematrix)
                    ctxI.set_source_rgba(c,c,c,c)
                    ctxI.mask(pattern)
                    drawn_regions.union(new_region)
                    break
    I.flush()
    I.write_to_png("wordle-cairo.png")
    Image.open("wordle-cairo.png").show()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
#
#  pycairo/cairocffi-based glyph-lcd example - Copyright 2017 Hin-Tak Leung
#  Distributed under the terms of the new BSD license.
#
#  rewrite of the numply,matplotlib-based example from Nicolas P. Rougier
#  - Not immitating the upside-downness
#  - Do LCD_V side by side also
#
#  Limitation: Suface.get_data() is not in the "python 3, pycairo < 1.11" combo.
#
# -----------------------------------------------------------------------------
'''
Glyph bitmap LCD display (sub-pixel) rendering
'''
from freetype import *
# use Matrix() from Cairo instead of from Freetype
from cairo import Context, ImageSurface, FORMAT_ARGB32, SurfacePattern, FILTER_BEST, Matrix

if __name__ == '__main__':
    from PIL import Image
    from numpy import ndarray, ubyte, uint32

    face = Face('./Vera.ttf')
    face.set_char_size( 48*64 )
    face.load_char('S', FT_LOAD_RENDER |
                        FT_LOAD_TARGET_LCD )
    bitmap = face.glyph.bitmap
    width  = face.glyph.bitmap.width//3
    rows   = face.glyph.bitmap.rows
    pitch  = face.glyph.bitmap.pitch
    copybuffer = (c_ubyte * (pitch * rows))()
    memmove(pointer(copybuffer), bitmap._FT_Bitmap.buffer,
            pitch * rows)

    I = ImageSurface(FORMAT_ARGB32, width, rows)
    ndR = ndarray(shape=(rows,width), buffer=copybuffer,
                  dtype=ubyte, order='C',
                  offset=0,
                  strides=[pitch, 3])
    ndG = ndarray(shape=(rows,width), buffer=copybuffer,
                  dtype=ubyte, order='C',
                  offset=1,
                  strides=[pitch, 3])
    ndB = ndarray(shape=(rows,width), buffer=copybuffer,
                  dtype=ubyte, order='C',
                  offset=2,
                  strides=[pitch, 3])
    try:
        ndI = ndarray(shape=(rows,width), buffer=I.get_data(),
                      dtype=uint32, order='C',
                      strides=[I.get_stride(), 4])
    except NotImplementedError:
       raise SystemExit("For python 3.x, you need pycairo >= 1.11+ (from https://github.com/pygobject/pycairo)")
    # 255 * 2**24 = opaque
    ndI[:,:] = 255 * 2**24 + ndR[:,:] * 2**16 + ndG[:,:] * 2**8 + ndB[:,:]
    I.mark_dirty()

    surface = ImageSurface(FORMAT_ARGB32, 800, 600)
    ctx = Context(surface)

    ctx.set_source_surface(I, 0, 0)
    pattern = ctx.get_source()
    SurfacePattern.set_filter(pattern, FILTER_BEST)
    scale = 480.0 / rows
    scalematrix = Matrix()
    scalematrix.scale(1.0/scale,1.0/scale)
    scalematrix.translate(-(400.0 - width *scale /2.0 )+200, -60)
    pattern.set_matrix(scalematrix)
    ctx.paint()

    # we need this later for shifting the taller LCD_V glyph up.
    rows_old = rows

    # LCD_V
    face.load_char('S', FT_LOAD_RENDER |
                        FT_LOAD_TARGET_LCD_V )
    bitmap = face.glyph.bitmap
    width  = face.glyph.bitmap.width
    rows   = face.glyph.bitmap.rows//3
    pitch  = face.glyph.bitmap.pitch
    copybuffer = (c_ubyte * (pitch * face.glyph.bitmap.rows))()
    memmove(pointer(copybuffer), bitmap._FT_Bitmap.buffer,
            pitch * face.glyph.bitmap.rows)

    I = ImageSurface(FORMAT_ARGB32, width, rows)
    ndR = ndarray(shape=(rows,width), buffer=copybuffer,
                  dtype=ubyte, order='C',
                  offset=0,
                  strides=[pitch*3, 1])
    ndG = ndarray(shape=(rows,width), buffer=copybuffer,
                  dtype=ubyte, order='C',
                  offset=pitch,
                  strides=[pitch*3, 1])
    ndB = ndarray(shape=(rows,width), buffer=copybuffer,
                  dtype=ubyte, order='C',
                  offset=2*pitch,
                  strides=[pitch*3, 1])
    ndI = ndarray(shape=(rows,width), buffer=I.get_data(),
                  dtype=uint32, order='C',
                  strides=[I.get_stride(), 4])
    # 255 * 2**24 = opaque
    ndI[:,:] = 255 * 2**24 + ndR[:,:] * 2**16 + ndG[:,:] * 2**8 + ndB[:,:]
    I.mark_dirty()

    ctx.set_source_surface(I, 0, 0)
    pattern = ctx.get_source()
    SurfacePattern.set_filter(pattern, FILTER_BEST)
    # Re-use the sane scale (LVD_V taller than LCD),
    # but adjust vertical position to line up mid-height of glyphs
    adjust = (rows - rows_old) * 240.0 /rows
    scalematrix.translate(-400, adjust)
    pattern.set_matrix(scalematrix)
    ctx.paint()
    #
    surface.flush()
    surface.write_to_png("glyph-lcd-cairo.png")
    Image.open("glyph-lcd-cairo.png").show()
